/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef INTERFACE_H
#define INTERFACE_H

#include "Math.h"

namespace SharpConstruct
{
	namespace Interface
	{
		const int CONTROL_HEIGHT = 20;
		const int SLIDER_WIDTH = 200;
		const int TEXT_HEIGHT = 15;
		const int TEXT_WIDTH = 8;
		
		class MouseData
		{
		public:
			static const float MOUSE_DOWN_REPEAT_DELAY = 0.1;	
		
			MouseData() : Location( 0, 0 ), DownLocation( 0, 0 ),
		                  LeftButtonDown( false ), MiddleButtonDown( false ),
			              RightButtonDown( false ), S1ButtonDown( false ),
			              S2ButtonDown( false )
			{}
				
			bool AnyButtonDown()
			{
				if( LeftButtonDown || MiddleButtonDown || RightButtonDown ||
				    S1ButtonDown || S2ButtonDown )
					return true;
				else
					return false;
			}
		
			Math::Point2D< int > Location;
			Math::Point2D< int > DownLocation;
			bool LeftButtonDown;
			bool MiddleButtonDown;
			bool RightButtonDown;
			bool S1ButtonDown;
			bool S2ButtonDown;
		};
		
		class KeyData
		{
		public:
			KeyData() : VOID_KEY( 0xffffff ), Key( 0 ), ShiftDown( false ),
			            ControlDown( false ), AltDown( false )
			{}
			KeyData( int k, bool s = false, bool c = false, bool a = false )
			: VOID_KEY( 0xffffff ), Key( k ), ShiftDown( s ), ControlDown( c ),
			  AltDown( a )
			{}
				
			char VOID_KEY; // Would like it to be const...
				
			bool operator==( KeyData t )
			{
				if( Key == t.Key && ShiftDown == t.ShiftDown &&
				    ControlDown == t.ControlDown && AltDown == t.AltDown )
					return true;
				else
					return false;
			}
			
			char Key;
			bool ShiftDown;
			bool ControlDown;
			bool AltDown;
			bool Enter;
			bool Delete;
			bool Backspace;
			bool LeftArrow;
			bool RightArrow;
			bool Home;
			bool End;
			bool Tab;
		};
		
		class Event
		{
		public:
			static MouseData Mouse;
			static KeyData Keyboard;
		};
		
		/* All classes which draw to the screen should inherit from Drawable.
		   Drawable keeps track of the current display, as well as the z-order
		   and whether a widget is visible or not.
		*/
		class Drawable
		{
		public:
			inline Drawable()
			: _order( 0 ), _visible( true )
			{}
			
			static inline void ResetOrderCount()
			{
				_order_count = 0;
			}
			inline int Order() const
			{
				return _order;
			}
			inline bool Visible() const
			{
				return _visible;
			}
			inline void SetVisible( bool v )
			{
				_visible = v;
			}
			virtual void Hide() = 0;
		
			virtual void Draw();
		protected:
			virtual void _draw();
		private:
			static int _order_count;
			int _order;
			bool _visible;
		};
		
		class Handler
		{
		public:
			virtual ~Handler()
			{}
		
			virtual void MouseDown() = 0;
			virtual void MouseStillDown() = 0;
			virtual void MouseUp() = 0;
			virtual void MouseMove() = 0;
			virtual void ScrollDown() = 0;
			virtual void ScrollUp() = 0;
			virtual void KeyDown() = 0;
		};
	}
}
#endif // INTERFACE_H
