/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef MESH_H
#define MESH_H

#include "Brush.h"
#include "Optimized.h"
#include <gts.h>
#include <map>
#include <set>
#include <sigc++/sigc++.h>
#include <string>
#include <vector>

namespace SharpConstruct
{
	namespace Interface
	{
		class ViewGL;
	}
	class ExternalRender;
}

/** Stores and manipulates vertex and polygon data.
 *
 * Provides storage for vertex and polygon data, as well as data used to
 * display the mesh, and data relating the vertex and polygon data. Active data
 * is kept involving the area of the mesh currently being edited, and the
 * edit configuration. Functions are provided for file input/output, interactive
 * editing, deformations, and retrieval of mesh information.
 *
 * Note that in the context of this class, "selection" refers to the sphere of
 * points centered at _center and within the diameter of _diameter. "Area"
 * refers to whatever mirror of the selection is being worked on. The area may
 * be the selection when not mirrored over any axis.
**/
class Mesh
{
	friend class SharpConstruct::Interface::ViewGL;
	friend class SharpConstruct::ExternalRender;
	
public:
	Mesh();
	~Mesh()
	{}

	// Basic file input/output control
	void Load( std::string filename );
	void Save( std::string filename );

	// Create model data
	void BecomePlane( int, int );
	void BecomeSphere( int );

	// Brushing Operations
	void DisplaceAreas();
	void StartGrab();
	void ContinueGrab( float, float );
	void EndGrab();
	void SmoothAreas();
	void MaskAreas();

	// Deformations
	void Move( float, float, float );
	void Resize( float, float, float );
	void Smooth( float );
	void Flood( float );
	void Subdivide();
	void Optimize();
	int RemoveDuplicates();
	void Mirror( bool, bool, bool );
	void Unify();

	//void SetTexture( Magick::Image * );

	// Mesh information access
	sigc::signal< void >& Changed();
	std::string FileName() const;
	bool Saved() const;
	std::string GetModelStatisticsString();
	float GetDiameter();
	void CopyEssentials( Mesh & );
	void ClearExtraData();
	
	static SharpConstruct::Brush *MainBrush;
private:
	// GTS functions and temporary data
	void _to_gts_surface( GtsSurface & );
	void _from_gts_surface( GtsSurface & );
	static void _add_gts_vertex( GtsColorVertex *, gpointer * );
	static void _add_gts_triangle( GtsTriangle *, gpointer * );
	static GtsVertex *_gts_segment_midvertex( GtsSegment *, GtsVertexClass * );
	static gboolean _stop_gts_coarsen( gdouble, guint, gpointer );
	static gboolean _stop_gts_subdivide( gdouble, guint, gpointer );
	std::map< GtsColorVertex *, int > _vertex_map;

	// File input/output
	void _load_obj( std::string );
	void _save_obj( std::string );
	void _save_wrl( std::string );

	void _displace_area( SharpConstruct::Optimized::Point3D, int );
	void _smooth_area( SharpConstruct::Optimized::Point3D, int );
	void _continue_grab( std::vector< int > &, std::vector< float > &,
	                     SharpConstruct::Optimized::Normal3D );
	void _mask_areas( SharpConstruct::Optimized::Point3D, int );

	void _calculate_bounding_sphere();
	void _find_active_vertices( std::vector< int > &,
	                            std::vector< float > &,
								SharpConstruct::Optimized::Point3D, int );
	void _is_vertex_active( unsigned, std::set< int > &,
	                        std::vector< int > &, std::vector< float > &,
	                        SharpConstruct::Optimized::Point3D );
	SharpConstruct::Optimized::Point3D _get_mirror_point( bool, bool, bool );

	// Calculate normals
	void _flip_normals();
	SharpConstruct::Optimized::Point3D _calculate_area_normal();
	void _update_polygon_normal( int );
	void _recalculate_all_normals();
	void _recalculate_modified_normals();
	
	// Object data
	void _change_operation();
	sigc::signal< void > _changed;
	bool _saved;
	std::string _filename;
	
	SharpConstruct::Optimized::Point3DVector _vertex_locations;
	SharpConstruct::Optimized::Point3DVector _vertex_normals;
	std::vector< SharpConstruct::Color > _vertex_colors;
	std::vector< unsigned > _indices;
	// Stores the normals of each polygon
	SharpConstruct::Optimized::Point3DVector _polygon_normals;
	// Stores the polygons that use each vertex
	std::vector< std::vector< int > > _vertex_polygon_indices;
	
	std::list< unsigned > _damaged_polygon_indices;
	
	// Active data
	static SharpConstruct::Optimized::Point3D _current_selection;
	unsigned _selected_vertex[ 8 ];
	std::vector< int > _active_vertex_indices;
	std::vector< float > _active_vertex_distances;
	static SharpConstruct::Optimized::Normal3D _view_up_normal;
	static SharpConstruct::Optimized::Normal3D _view_right_normal;
	unsigned _lowest_index, _highest_index;
	std::vector< int > _modified_vertex_indices;
	std::vector< int > _grab_vertices[ 8 ];
	std::vector< float > _grab_vertex_distances[ 8 ];
	//static Magick::Image *_texture;
	
	// Bounding Sphere
	float _diameter;
	SharpConstruct::Optimized::Point3D _center;
};

#endif // MESH_H
