/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "ColorPicker.h"
#include "TopWindow.h"
#include <GL/gl.h>
#include <limits>

/* This file contains some of the least-readable code in SharpConstruct.
 * Anyone wishing to rectify the situation...
 */

using namespace SharpConstruct::Interface;

ColorSwatch::ColorSwatch()
{
	SetColor( Color( 1, 1, 1 ) );
}

void ColorSwatch::SetColor( const Color& c )
{
	_color = c;
}

void ColorSwatch::_draw()
{
	glColor3f( 0.3, 0.3, 0.3 );
	glRecti( X(), Y(), Right() + 1, Bottom() + 1 );
	glColor3f( _color.Red, _color.Green, _color.Blue );
	glRecti( X() + 1, Y() + 1, Right(), Bottom() );
}

ValueIntensityPicker::ValueIntensityPicker()
: _base_color( 1, 0, 0 ), _final_color( 1, 0, 0 ), _selection( 0, 0 )
{
	MouseDownSignal.connect( sigc::mem_fun( *this, &ValueIntensityPicker::_set ) );
	MouseMoveSignal.connect( sigc::mem_fun( *this, &ValueIntensityPicker::_set ) );
}

SharpConstruct::Color ValueIntensityPicker::FinalColor() const
{
	return _final_color;
}
void ValueIntensityPicker::SetFinalColor( Color cf )
{
	{
		Color c = cf;
		const int b = static_cast< int >( round( W() / 8.0f ) );
		const float min = std::min( std::min( c.Red, c.Green ), c.Blue );
		c.Red -= min;
		c.Green -= min;
		c.Blue -= min;
		const float mlt = 1 / std::max( std::max( c.Red, c.Green ), c.Blue );
		c.Red *= mlt;
		c.Green *= mlt;
		c.Blue *= mlt;
	
		const float br = 1 - ( ( cf.Red + cf.Green + cf.Blue ) / 3.0 );
		_selection.SetY( br * ( H() - 6 ) );
	
		// For saturation, we know that the range is [0,2]
		const float sat_range = 2.0f;
		const float sat = fabs( cf.Red - cf.Green ) +
		                  fabs( cf.Green - cf.Blue ) +
		                  fabs( cf.Red - cf.Blue );
		
		_selection.SetX( ( sat * ( W() - 6 ) ) / sat_range );
	}
	
	ChangedSignal( _final_color );
}

void ValueIntensityPicker::SetBaseColor( const Color& c )
{
	_base_color = c;
	_draw();
	ChangedSignal( _final_color );
}

void ValueIntensityPicker::_draw()
{
	glBegin( GL_TRIANGLE_FAN );
	glColor3f( _base_color.Red, _base_color.Green, _base_color.Blue );
	glVertex2f( Right(), Center().Y() );
	glColor3f( 1, 1, 1 );
	glVertex2f( Right(), Y() );
	glColor3f( 1, 1, 1 );
	glVertex2f( X(), Y() );
	glColor3f( 0, 0, 0 );
	glVertex2f( X(), Bottom() );
	glColor3f( 0, 0, 0 );
	glVertex2f( Right(), Bottom() );
	glEnd();

	glReadPixels( _selection.X() + X(),
	              TopWindow::Display().H() - ( _selection.Y() + Y() + 1 ),
	              1, 1, GL_RGB, GL_FLOAT, &_final_color.Red );
	
	{
		glBegin( GL_LINE_STRIP );
		glColor3f( 0, 0, 0 );
		glVertex2f( X() + _selection.X(), Y() + _selection.Y() );
		glVertex2f( X() + _selection.X() + 6, Y() + _selection.Y() );
		glVertex2f( X() + _selection.X() + 6, Y() + _selection.Y() + 6 );
		glVertex2f( X() + _selection.X(), Y() + _selection.Y() + 6 );
		glVertex2f( X() + _selection.X(), Y() + _selection.Y() );
		glColor3f( 1, 1, 1 );
		glVertex2f( X() + _selection.X() + 1, Y() + _selection.Y() + 1 );
		glVertex2f( X() + _selection.X() + 5, Y() + _selection.Y() + 1 );
		glVertex2f( X() + _selection.X() + 5, Y() + _selection.Y() + 5 );
		glVertex2f( X() + _selection.X() + 1, Y() + _selection.Y() + 5 );
		glVertex2f( X() + _selection.X() + 1, Y() + _selection.Y() + 1 );
		glEnd();
	}
}

void ValueIntensityPicker::_set()
{
	if( Event::Mouse.LeftButtonDown )
	{
		const int cursor_offset = 3;

		_selection = Event::Mouse.Location;
		_selection.SetX( _selection.X() - ( X() + cursor_offset ) );
		_selection.SetY( _selection.Y() - ( Y() + cursor_offset ) );
		
		if( _selection.X() < 0 )
			_selection.SetX( 0 );
		else if( _selection.X() > W() - 8 )
			_selection.SetX( W() - 8 );
		if( _selection.Y() < 0 )
			_selection.SetY( 0 );
		else if( _selection.Y() > H() - 8 )
			_selection.SetY( H() - 8 );
		
		_draw();

		ChangedSignal( _final_color );
	}
}

HuePicker::HuePicker()
: _selection( 0, 0 )
{
	MouseDownSignal.connect( sigc::mem_fun( *this, &HuePicker::_set ) );
	MouseMoveSignal.connect( sigc::mem_fun( *this, &HuePicker::_set ) );
}

void HuePicker::SetColor( Color c )
{
	if( c.IsGrey() )
	{
		_selection.SetX( 0 );
		_selection.SetY( 0 );
		c = Color( 1, 0, 0 );
	}
	else
	{
		/* Brought to you by the association of Late-Night Coders.
		 *
		 * This code is not to be considered bug free, readable, accurate, or
		 * in any way useful.
		 *
		 * And that's putting it nicely.
		 */
		
		const int b = static_cast< int >( round( W() / 8.0f ) );
		const float min = std::min< float >( std::min< float >( c.Red, c.Green ), c.Blue );

		c.Red -= min;
		c.Green -= min;
		c.Blue -= min;
	
		const float mlt = 1.0 / std::max< float >( std::max< float >( c.Red, c.Green ), c.Blue );
		c.Red *= mlt;
		c.Green *= mlt;
		c.Blue *= mlt;
		
		const float e = std::numeric_limits< float >::epsilon();
		
		const float W2d3 = 2.0 / 3.0 * ( W() - b );
		if( c.Red > 1 - e && c.Blue < e )
		{
			_selection.SetX( W2d3 * c.Green );
			_selection.SetY( 0 );
		}
		else if( c.Green > 1 - e && c.Blue < e )
		{
			// Split at the top-left corner
			c.Red = 1 - c.Red;
			if( c.Red < 0.5 )
			{
				_selection.SetX( static_cast< int >( W2d3 + W2d3 * c.Red ) );
				_selection.SetY( 0 );
			}
			else
			{
				_selection.SetX( W() - b );
				_selection.SetY( static_cast< int >( W2d3 * c.Red - W2d3 / 2 ) );
			}
			c.Red = 1 - c.Red;
		}
		else if( c.Red < e && c.Green > 1 - e )
		{
			_selection.SetX( W() - b );
			_selection.SetY( static_cast< int >( W2d3 / 2 + W2d3 * c.Blue ) );
		}
		else if( c.Red < e && c.Blue > 1 - e )
		{
			_selection.SetX( static_cast< int >( W2d3 / 2 + W2d3 * c.Green ) );
			_selection.SetY( H() - b );
		}
		else if( c.Green < e && c.Blue > 1 - e )
		{
			c.Red = 1 - c.Red;
			if( c.Red < 0.5 )
			{
				_selection.SetX( 0 );
				_selection.SetY( static_cast< int >( W2d3 + W2d3 * c.Red ) );
			}
			else
			{
				_selection.SetX( static_cast< int >( W2d3 * c.Red - W2d3 / 2 ) );
				_selection.SetY( H() - b );
			}
			c.Red = 1 - c.Red;
		}
		else if( c.Red > 1 - e && c.Green < e )
		{
			_selection.SetX( 0 );
			_selection.SetY( static_cast< int >( W2d3 * c.Blue ) );
		}
	}
	_color = c;
	ChangedSignal( _color );
}

void HuePicker::_draw()
{
	const float b = W() / 8.0f;
	
	// Border
	glColor3f( 0.3, 0.3, 0.3 );
	glRecti( X(), Y(), Right() + 1, Bottom() + 1 );
	
	// Draw the band of colors - red, yellow, green, cyan, blue, magenta
	glBegin( GL_QUAD_STRIP );
	{
		const int left = X() + 1;
		glColor3f( 1, 0, 0 );
		glVertex2f( left, Y() );
		glVertex2f( left + b, Y() + b );
		glColor3f( 1, 1, 0 );
		glVertex2f( left + W() * ( 2.0f / 3 ), Y() );
		glVertex2f( left + W() * ( 2.0f / 3 ), Y() + b );
		glColor3f( 0.5, 1, 0 );
		glVertex2f( Right(), Y() );
		glVertex2f( Right() - b, Y() + b );
		glColor3f( 0, 1, 0 );
		glVertex2f( Right(), Y() + H() * ( 1.0f / 3 ) );
		glVertex2f( Right() - b, Y() + H() * ( 1.0f / 3 ) );
		glColor3f( 0, 1, 1 );
		glVertex2f( Right(), Bottom() );
		glVertex2f( Right() - b, Bottom() - b );
		glColor3f( 0, 0, 1 );
		glVertex2f( left + W() * ( 1.0f / 3 ), Bottom() );
		glVertex2f( left + W() * ( 1.0f / 3 ), Bottom() - b );
		glColor3f( 0.5, 0, 1 );
		glVertex2f( left, Bottom() );
		glVertex2f( left + b, Bottom() - b );
		glColor3f( 1, 0, 1 );
		glVertex2f( left, Y() + H() * ( 2.0f / 3 ) );
		glVertex2f( left + b, Y() + H() * ( 2.0f / 3 ) );
		glColor3f( 1, 0, 0 );
		glVertex2f( left, Y() );
		glVertex2f( left + b, Y() + b );
	} glEnd();
	
	// Read in current color from the selector
	glReadPixels( _selection.X() + X() + 3,
	              TopWindow::Display().H() - ( _selection.Y() + Y() + 3 ),
	              1, 1, GL_RGB, GL_FLOAT, &_color.Red );
	
	// Draw selector region
	glBegin( GL_LINE_STRIP );
	{		
		glColor3f( 0, 0, 0 );
		glVertex2f( X() + _selection.X(), Y() + _selection.Y() );
		glVertex2f( X() + _selection.X() + b, Y() + _selection.Y() );
		glVertex2f( X() + _selection.X() + b, Y() + _selection.Y() + b );
		glVertex2f( X() + _selection.X(), Y() + _selection.Y() + b );
		glVertex2f( X() + _selection.X(), Y() + _selection.Y() );
		glColor3f( 1, 1, 1 );
		glVertex2f( X() + _selection.X() + 1, Y() + _selection.Y() + 1 );
		glVertex2f( X() + _selection.X() + b - 1, Y() + _selection.Y() + 1 );
		glVertex2f( X() + _selection.X() + b - 1, Y() + _selection.Y() + b - 1 );
		glVertex2f( X() + _selection.X() + 1, Y() + _selection.Y() + b - 1 );
		glVertex2f( X() + _selection.X() + 1, Y() + _selection.Y() + 1 );
	} glEnd();
}

void HuePicker::_set()
{
	if( Event::Mouse.LeftButtonDown )
	{
		const float b = W() / 8.0f;
		
		_selection.SetX( Event::Mouse.Location.X() - ( X() + b / 2 ) );
		_selection.SetY( Event::Mouse.Location.Y() - ( Y() + b / 2 ) );
		
		if( _selection.X() < 0 )
			_selection.SetX( 0 );
		else if( _selection.X() > W() - b )
			_selection.SetX( W() - b );
		if( _selection.Y() < 0 )
			_selection.SetY( 0 );
		else if( _selection.Y() > H() - b )
			_selection.SetY( H() - b );
		
		if( _selection.Y() < _selection.X() )
		{
			if( _selection.Y() < W() - _selection.X() - b )
				_selection.SetY( 0 );
			else
				_selection.SetX( W() - b );
		}
		else
		{
			if( _selection.Y() < W() - _selection.X() - b )
				_selection.SetX( 0 );
			else
				_selection.SetY( H() - b );
		}

		_draw();
		ChangedSignal( _color );
	}
}

ColorPicker::ColorPicker( int h )
: Group( h + 20, h, "COLORPICKER" )
{
	const float border = h / 8.0f;

	_hue_picker.SetSize( h, h );
	_hue_picker.ChangedSignal.connect( sigc::mem_fun( _vi_picker, &ValueIntensityPicker::SetBaseColor ) );

	_vi_picker.Set( border, border, border * 6, border * 6 );
	_vi_picker.ChangedSignal.connect( ChangedSignal.make_slot() );
	_vi_picker.ChangedSignal.connect( sigc::mem_fun( _swatch, &ColorSwatch::SetColor ) );

	_swatch.SetSize( 20, h );
	_swatch.SetX( _hue_picker.Right() );
	
	PushBack( _hue_picker );
	PushBack( _vi_picker );
	PushBack( _swatch );
}

SharpConstruct::Color ColorPicker::FinalColor() const
{
	return _vi_picker.FinalColor();
}

void ColorPicker::SetColor( const Color& c )
{
	_hue_picker.SetColor( c );
	_vi_picker.SetFinalColor( c );
	_swatch.SetColor( c );
}
