/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "DeformDialog.h"
#include "MainWindow.h"

using namespace SharpConstruct::Interface;

DeformDialog::DeformDialog( MainWindow& mw )
: Dialog( "Deform" ),
  _parent( mw ),
  _dynamic( 2 ),
  _axes( false, 0, 4 ), _xaxis( "X" ), _yaxis( "Y" ), _zaxis( "Z" ),
  _move( "Move" ),
  _scale( "Scale" ),
  _smooth( "Smooth" ),
  _flood( "Flood" ),
  _subdivide( "Subdivide" ),
  _optimize( "Optimize" ),
  _mirror( "Mirror" ),
  _unify( "Unify" )
{
	// Yes, I know...
	_xaxis.SetW( 65 );
	_yaxis.SetW( 66 );
	_zaxis.SetW( 65 );
	
	_move.SetProperties( -100, 100, 1, 0, false );
	_scale.SetProperties( -100, 100, 1, 0, false );
	_smooth.SetProperties( -100, 100, 1, 0, false );
	_flood.SetProperties( 0, 100, 1, 0, false );
	
	_axes.PushBack( _xaxis );
	_axes.PushBack( _yaxis );
	_axes.PushBack( _zaxis );
	PushBack( _axes );
	PushBack( _move );
	PushBack( _scale );
	PushBack( _smooth );
	PushBack( _flood );
	PushBack( _subdivide );
	PushBack( _optimize );
	PushBack( _mirror );
	PushBack( _unify );
	
	_move.MouseDownSignal.connect( sigc::mem_fun( *this, &DeformDialog::_switch_history ) );
	_move.ChangedSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_move ) );
	_move.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_revert_history ) );
	_move.SetSignal.connect( sigc::mem_fun( *this, &DeformDialog::_transfer ) );
	
	_scale.MouseDownSignal.connect( sigc::mem_fun( *this, &DeformDialog::_switch_history ) );
	_scale.ChangedSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_scale ) );
	_scale.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_revert_history ) );
	_scale.SetSignal.connect( sigc::mem_fun( *this, &DeformDialog::_transfer ) );
	
	_smooth.MouseDownSignal.connect( sigc::mem_fun( *this, &DeformDialog::_switch_history ) );
	_smooth.ChangedSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_smooth ) );
	_smooth.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_revert_history ) );
	_smooth.SetSignal.connect( sigc::mem_fun( *this, &DeformDialog::_transfer ) );
	
	_flood.MouseDownSignal.connect( sigc::mem_fun( *this, &DeformDialog::_switch_history ) );
	_flood.ChangedSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_flood ) );
	_flood.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_revert_history ) );
	_flood.SetSignal.connect( sigc::mem_fun( *this, &DeformDialog::_transfer ) );
	
	_subdivide.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_subdivide ) );
	_optimize.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_optimize ) );
	_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_mirror ) );
	_unify.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_unify ) );
	
	SetX( 0 );
	SetY( 212 );
}

void DeformDialog::_switch_history()
{
	_old_history = _parent.View()._history;
	_dynamic.Clear();
	_dynamic.GetCurrentMesh().CopyEssentials( _old_history->GetCurrentMesh() );
	_parent.View()._history = &_dynamic;
}
void DeformDialog::_revert_history()
{
	_parent.View()._history = _old_history;
	
}
void DeformDialog::_transfer()
{
	_old_history->AddMesh();
	_old_history->GetCurrentMesh().CopyEssentials( _dynamic.GetCurrentMesh() );
}

void DeformDialog::_cb_move( float v )
{
	_dynamic.Undo();
	_dynamic.AddMesh();
	float scaled_value = v / 100;
	scaled_value *= _dynamic.GetCurrentMesh().GetDiameter();
	_dynamic.GetCurrentMesh().Move( scaled_value * _xaxis.Value(),
	                                scaled_value * _yaxis.Value(),
	                                scaled_value * _zaxis.Value() );
	_parent.View().Refresh();
}
void DeformDialog::_cb_scale( float v )
{
	_dynamic.Undo();
	_dynamic.AddMesh();
	float clipped_value = v / 50;
	if( clipped_value > 0 )
		clipped_value++;
	else
		clipped_value--;
	_dynamic.GetCurrentMesh().Resize( clipped_value * _xaxis.Value(),
	                                  clipped_value * _yaxis.Value(),
	                                  clipped_value * _zaxis.Value() );
	_parent.View().Refresh();
}
void DeformDialog::_cb_smooth( float v )
{
	_dynamic.Undo();
	_dynamic.AddMesh();
	_dynamic.GetCurrentMesh().Smooth( v / 100 );
	_parent.View().Refresh();
}
void DeformDialog::_cb_flood( float v )
{
	_dynamic.Undo();
	_dynamic.AddMesh();
	_dynamic.GetCurrentMesh().Flood( v / 100 );
	_parent.View().Refresh();
}
void DeformDialog::_cb_subdivide()
{
	_parent.SetStatusText( "Subdividing..." );
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Subdivide();
	_parent.SetStatusText( "Subdivide complete; " + _parent.GetMeshHistory().
		GetCurrentMesh().GetModelStatisticsString() );
}
void DeformDialog::_cb_optimize()
{
	_parent.SetStatusText( "Optimizing..." );
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Optimize();
	_parent.SetStatusText( "Optimize complete; " + _parent.GetMeshHistory().
		GetCurrentMesh().GetModelStatisticsString() );
}
void DeformDialog::_cb_mirror()
{
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Mirror( _xaxis.Value(), _yaxis.Value(),
	                                         _zaxis.Value() );
}
void DeformDialog::_cb_unify()
{
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Unify();
	//Refresh( true );
}
