/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Dialog.h"
#include "Text.h"
#include "TopWindow.h"
#include <GL/gl.h>

#include <iostream>

using namespace SharpConstruct::Interface;

Dialog::Dialog( std::string lbl )
: _title( "" ), _close( "X" )
{
	SetBorder( 4 );
	SetInternalBorder( 4 );

	_close.SetH( 24 );
	_close.MouseUpSignal.connect( sigc::retype_return< void >( sigc::mem_fun( *this, &Dialog::Close ) ) );
	_close_area.SetCorporeal( false );
	_close_area.PushBack( _close );
	
	SetW( 100 );
	SetH( 100 );
	SetLabel( lbl );
	_title.SetH( 16 );
	_title.SetVisible( false );
	PushBack( _title );
	PushBack( _close_area );
	MouseDownSignal.connect( sigc::mem_fun( *this, &Dialog::_mouse_down ) );
	MouseMoveSignal.connect( sigc::mem_fun( *this, &Dialog::_mouse_move ) );
}

void Dialog::Reorganize()
{
	VerticalGroup::Reorganize();
	_close.SetX( Right() - _close.W() );
	_close.SetY( Y() );
	_title.SetRight( _close.X() - 10 );
}

void Dialog::DoEvents()
{
	//
}
bool Dialog::Close()
{
	SetVisible( false );
	return true;
}

void Dialog::Show()
{
	Widget::Show();
	RaiseSignal( this );
}

void Dialog::_draw()
{
	// Shadow...
	glBegin( GL_QUADS );
	{
		glColor4f( 0, 0, 0, 0.5 );
		glVertex2f( Right(), Y() );
		glVertex2f( Right(), Bottom() );
		glColor4f( 0, 0, 0.5, 0 );
		glVertex2f( Right() + 6, Bottom() + 6 );
		glVertex2f( Right() + 6, Y() + 6 );
		
		glColor4f( 0, 0, 0, 0.5 );
		glVertex2f( X(), Bottom() );
		glVertex2f( Right(), Bottom() );
		glColor4f( 0, 0, 0, 0 );
		glVertex2f( Right() + 6, Bottom() + 6 );
		glVertex2f( X() + 6, Bottom() + 6 );
	} glEnd();
	
	Group::_draw();
	
	glBegin( GL_QUADS );
	{
		glColor4f( 0.64, 0.74, 0.84, 0 );
		glVertex2f( X() + 1, Y() + 1 );
		glVertex2f( Right(), Y() + 1 );
		glColor4f( 0.64, 0.74, 0.84, 1 );
		glVertex2f( Right(), Y() + 24 );
		glVertex2f( X() + 1, Y() + 24 );
	} glEnd();
	GLOut::glout << GLFormat::Align::Center()
	             << GLFormat::Location( X() + W() / 2, Y() + 12 ) << Label();
	
	// Redraw this above the title bar...
	_close.Draw();
}

void Dialog::Cancel()
{
	CancelSignal();
	SetVisible( false );
}

void Dialog::MouseDown()
{	
	if( Visible() && Enabled() && Contains( Event::Mouse.Location ) )
	{
		ResponderWidgets::Responders.PushBack( this );
		ResponderDialogs::Responders.PushBack( this );
		for( Iterator i = Begin(); i != End(); i++ )
			( *i )->MouseDown();
	}
}

void Dialog::_mouse_down()
{
	_relative_click_location.Set( Event::Mouse.Location.X() - X(),
	                              Event::Mouse.Location.Y() - Y() );
	RaiseSignal.emit( this );
}

void Dialog::_mouse_move()
{	
	if( Event::Mouse.LeftButtonDown )
	{
		SetX( Event::Mouse.Location.X() - _relative_click_location.X() );
		SetY( Event::Mouse.Location.Y() - _relative_click_location.Y() );
		if( X() < TopWindow::Display().X() )
			SetX( TopWindow::Display().X() );
		if( Y() < TopWindow::Display().Y() )
			SetY( TopWindow::Display().Y() );
		if( Right() > TopWindow::Display().Right() )
			SetRight( TopWindow::Display().Right() );
		if( Y() > TopWindow::Display().Bottom() - 24)
			SetY( TopWindow::Display().Bottom() - 24 );
		//    Nasty (Group?) bug, things fly backwards once past edge...
		/*if( Right() > TopWindow::Display().Right() )
			SetRight( TopWindow::Display().Right() );
		if( Bottom() > TopWindow::Display().Bottom() )
			SetBottom( TopWindow::Display().Bottom() );*/
	}
}

DialogList::DialogList()
{
	SetCorporeal( false );
}

void DialogList::PushBack( Dialog& d )
{
	PushBack( &d );
}
void DialogList::PushBack( Dialog* d )
{
	d->RaiseSignal.connect( sigc::mem_fun( *this, &DialogList::_raise ) );
	Group::PushBack( d );
}

void DialogList::_raise( Dialog* d )
{
	Erase( Find( d ) );
	Group::PushBack( d );
}
