/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "FileDialog.h"
#include "InputDialog.h"
#include "MessageDialog.h"
#include "Text.h"
#include "TopWindow.h"
#include <GL/gl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>
#include <cstdlib>

using namespace SharpConstruct::Interface;

FileDialog::FileDialog( std::string label )
: Dialog( label ),
  _DEFAULT_HEIGHT( 300 ),
  //_all( true, 0, 4 ),
  //_dir_controls( new HorizontalGroup() ),
  _root( 0, 20, 40, 20, "Root" ),
  _home( 40, 20, 40, 20, "Home" ),
  _new_dir( 80, 20, 64, 20, "New Dir" ),
  _hidden( 144, 20, 96, 20, "Show hidden" ),
  _main( false, 0, 2 ),
  _dirs_list_group( true, 0, 0 ),
  _dirs( 0, 64, 300, _DEFAULT_HEIGHT - 86 ),
  _dir_up( 0, 44, 200, 20, "Dir Up" ),
  _dir_file_split( 204, 44, 8, _DEFAULT_HEIGHT - 66 ),
  _files( 204, 44, 300, _DEFAULT_HEIGHT - 66 ),
  _control_buttons( true, 4, 4 ),
  _cancel( 0, _DEFAULT_HEIGHT - 20, 56, 20, "Cancel" )
{
	_dir_file_split.ResizeSignal.connect( sigc::mem_fun( *this, &FileDialog::Reorganize ) );
	_hidden.MouseUpSignal.connect( sigc::mem_fun( *this, &FileDialog::RefreshEntries ) );
	_root.MouseUpSignal.connect( sigc::mem_fun( *this, &FileDialog::_go_root ) );
	_home.MouseUpSignal.connect( sigc::mem_fun( *this, &FileDialog::_go_home ) );
	_new_dir.MouseUpSignal.connect( sigc::mem_fun( *this, &FileDialog::_make_dir ) );
	_dirs.DoubleClickSignal().connect( sigc::mem_fun( *this, &FileDialog::_change_dir ) );
	_dir_up.MouseUpSignal.connect( sigc::mem_fun( *this, &FileDialog::_go_up ) );
	_cancel.MouseUpSignal.connect( sigc::mem_fun( *this, &Dialog::Cancel ) );
	_files.SelectSignal.connect( sigc::mem_fun( *this, &FileDialog::_set_file ) );
	_files.DoubleClickSignal().connect( sigc::mem_fun( *this, &FileDialog::_done ) );
	_path_entry.ReturnSignal().connect( sigc::mem_fun( *this, &FileDialog::_done ) );
	
	_dir_controls.PushBack( _root );
	_dir_controls.PushBack( _home );
	_dir_controls.PushBack( _new_dir );
	_dir_controls.PushBack( _hidden );
	
	_dirs_list_group.PushBack( _dir_up );
	_dirs_list_group.PushBack( _dirs );
	
	_main.PushBack( _dirs_list_group );
	_main.PushBack( _dir_file_split );
	_main.PushBack( _files );
	
	_control_buttons.PushBack( _cancel );
	
	
	PushBack( _dir_controls );
	PushBack( _main );
	PushBack( _path_entry );
	PushBack( _control_buttons );
	//PushBack( _cancel );
	
	//PushBack( _all );

	SetX( 212 );
	SetY( 88 );
	SetW( 700 );
	SetH( _DEFAULT_HEIGHT );
	
	_go_home();
	
	Reorganize();
}

void FileDialog::_change_dir()
{
	if( _dirs.Selection().length() )
		ChangeDir( _dirs.Selection() );
}
void FileDialog::ChangeDir( std::string p )
{
	_dir = Trailing( _dir );
	if( p[ 0 ] == '/' )
		_dir = p;
	else
		_dir = _dir + '/' + p;
	if( _dir.substr( 0, 2 ) == "//" )
		_dir.erase( 0, 1 );
	_dir = Trailing( _dir );
	_path_entry.SetLabel( _dir );
	RefreshEntries();
}

void FileDialog::RefreshEntries()
{
	int count;
	struct dirent** files;
	struct stat statinfo;
	std::string tmp;
	
	_dirs.std::vector< std::string >::clear();
	_files.std::vector< std::string >::clear();
	count = scandir( _dir.c_str(), &files, 0, alphasort );
	for( int i = 2; i < count; ++i ) // Start at two so we skip . and ..
	{
		if( !_hidden.Value() && files[ i ]->d_name[ 0 ] == '.' )
			continue;
		tmp = _dir + "/" + files[ i ]->d_name;
		stat( tmp.c_str(), &statinfo );
		if( S_ISDIR( statinfo.st_mode ) )
			_dirs.push_back( files[ i ]->d_name );
		else
			_files.push_back( files[ i ]->d_name );
	}
	free( files );
	
	_dirs.Refresh();
	_files.Refresh();
}

void FileDialog::Reorganize()
{
	Dialog::Reorganize();

	_dirs_list_group.SetRight( _dir_file_split.X() - 3 );
	_files.SetX( _dir_file_split.Right() + 3 );
	_dir_file_split.SetBounds( _main.X() + 4, _main.Right() - 4 );
	
	_dirs.Reorganize();
}

std::string FileDialog::Trailing( std::string path )
{
	int i = path.length() - 1;
	while( i >= 1 && path[ i ] == '/' )
		i--;

	return path.substr( 0, i + 1 );
}
std::string FileDialog::PathPart( std::string path )
{
	return path.substr( 0, path.rfind( '/' ) + 1 );
}
std::string FileDialog::FilePart( std::string path )
{
	return path.substr( path.rfind( '/' ) + 1 );
}
std::string FileDialog::HomePath()
{
	return getenv( "HOME" );
}

bool FileDialog::FileExists( std::string path )
{
	struct stat buffer;
	return !stat( path.c_str(), &buffer );
}

std::string FileDialog::GetPath()
{
	return _path_entry.Label();
}

void FileDialog::_done()
{
	Hide();
}

void FileDialog::_set_file( std::string file )
{
	_path_entry.SetLabel( _dir + "/" + file );
}

void FileDialog::_go_root()
{
	ChangeDir( "/" );
}

void FileDialog::_go_up()
{
	_dir = Trailing( _dir );
	ChangeDir( _dir.substr( 0, Trailing( _dir ).rfind( '/' ) + 1 ) );
}

void FileDialog::_go_home()
{
	ChangeDir( getenv( "HOME" ) );
}

void FileDialog::_make_dir()
{
	std::string ndir = InputDialog::ModalInput( TopWindow::Display(), "New Directory",
	                                            "Create Directory", "Cancel" );
	if( ndir.length() > 0 )
	{
		if( mkdir( ( _dir + "/" + ndir ).c_str(), 0755 ) == 0 )
			ChangeDir( ndir );
		else
		{
			std::string msg = "Cannot create directory '";
			msg += ndir + "': ";
			if( errno == EEXIST )
				msg += "File exists";
			else
				msg += "Unknown error";
			MessageDialog::ModalMessage( TopWindow::Display(), "New Directory",
										 msg, "Cancel" );
	
		}
	}
}
