/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "ListBox.h"
#include "Text.h"
#include "TopWindow.h"
#include <GL/gl.h>

#include <iostream>

using namespace SharpConstruct::Interface;

ListBox::ListBox( int x, int y, int w, int h )
: Group( 4 ), _ELEMENT_HEIGHT( 20 ), _bar( 0, 0, 20, h ), _selected( -1 )
{
	MouseDownSignal.connect( sigc::mem_fun( *this, &ListBox::_mouse_select ) );
	MouseMoveSignal.connect( sigc::mem_fun( *this, &ListBox::_mouse_select ) );
	MouseUpSignal.connect( sigc::mem_fun( *this, &ListBox::_mouse_up ) );
	ScrollDownSignal.connect( sigc::mem_fun( _bar, &Scrollbar::Increment ) );
	ScrollUpSignal.connect( sigc::mem_fun( _bar, &Scrollbar::Decrement ) );
	
	Group::PushBack( _bar );
	SetX( x );
	SetY( y );
	SetW( w );
	SetH( h );
}

void ListBox::MakeInactive()
{
	Widget::MakeInactive();
	if( !Contains() )
		MouseUpSignal.emit();
}

void ListBox::Refresh()
{
	_bar.SetMax( std::vector< std::string >::size() - H() / _ELEMENT_HEIGHT + 1 );
	_selected = -1;
}

std::string ListBox::Selection() const
{
	if( _selected == -1 )
		return "";
	else
		return std::vector< std::string >::at( _selected );
}

void ListBox::_draw()
{
	if( W() > 0 )
	{
		glEnable( GL_SCISSOR_TEST );
		glMatrixMode( GL_PROJECTION_MATRIX );
		glScissor( X(), TopWindow::Display().H() - Y() - H(), W(), H() );
		glMatrixMode( GL_MODELVIEW_MATRIX );
		Group::_draw();
		
		int y = _bar.Y() + 10;
		for( int i = _bar.Value(); i < std::vector< std::string >::size(); ++i )
		{
			if( i == _selected )
			{
				glColor4f( 0.24, 0.34, 0.44, 1 );
				glRecti( _bar.Right(), y - 10, Right(), y + 10 );
				glColor4f( 0.64, 0.74, 0.84, 1 );
				glRecti( _bar.Right() + 1, y - 9, Right(), y + 9 );
			}
			
			GLOut::glout << GLFormat::Align::Left()
						 << GLFormat::Location( _bar.Right() + 8, y )
						 << std::vector< std::string >::at( i );
			
			y += _ELEMENT_HEIGHT;
			if( y > Bottom() )
				break;
		}
		
		glDisable( GL_SCISSOR_TEST );
	}
}

void ListBox::push_back( const std::string s )
{
	std::vector< std::string >::push_back( s );
	Refresh();
}

void ListBox::_mouse_select()
{
	if( Event::Mouse.DownLocation.X() > _bar.Right() &&
		Event::Mouse.LeftButtonDown )
	{
		_selected = _bar.Value() + ( ( Event::Mouse.Location.Y() - Y() ) /
		                             ( float( Bottom() - Y() ) /
		                               ( float( H() ) /_ELEMENT_HEIGHT ) ) );
		
		if( _selected < 0 )
			_selected = 0;
		else if( _selected + 1 > std::vector< std::string >::size() )
			_selected = ((int)std::vector< std::string >::size())-1;
		
		if( std::vector< std::string >::size() > 0 )
		{
			// Scroll up to follow the selection
			while( _selected < _bar.Value() )
				_bar.SetValue( _bar.Value() - 1 );		
			
			// Scroll down to follow the selection
			while( _selected > _bar.Value() + (float)H() / _ELEMENT_HEIGHT - 1 )
				_bar.SetValue( _bar.Value() + 1 );
		}
	}
}

void ListBox::_mouse_up()
{
	if( Event::Mouse.Location.X() > _bar.Right() && _selected >= 0 )
		SelectSignal.emit( std::vector< std::string >::at( _selected ) );
}
