/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "InputDialog.h"
#include "MainWindow.h"
#include "MessageDialog.h"
#include "Preferences.h"
#include <GL/glu.h>
#include <sys/stat.h>
#include <iostream>

using namespace SharpConstruct::Interface;

MainWindow::MainWindow()
: TopWindow( 1024, 768, "SharpConstruct 0.07" ),
  _mesh_history( 2 ),
  _v( 100, 100, 640, 480 ),
  _deform_dialog( *this ),
  _new_dialog( *this ),
  _settings_dialog( *this ),
  _top_bar( false, 0, 4 ),
  
  _model( true, 4 ),
  
  _brush_settings( false, 4, 4 ),
  _brush_settings_push( true ),
 // _deform( 168, 200, "Deform" ),
  _texture( 168, 68, "Texture" ),
  //_render( 168, 120, "Render" ),
  _view( true, 4 ),
  _preferences( 168, 116, "Preferences" ),
  
  _load( 64, "Open" ),
  _new( 64, "New" ),
  _save( 64, "Save" ),
  _save_as( "Save As" ),
  _undo( 64, "Undo" ),
  _redo( 64, "Redo" ),

  _displace( "Draw" ),
  _grab( "Grab" ),
  _smooth( "Smooth" ),
  _deform( "Deform" ),
  _radius( "Radius" ),
  _strength( "Strength" ),
  _intensity( "Intensity" ),
  _add( "Add" ),
  _sub( "Sub" ),
  _rgb( "RGB" ),
  _x_mirror( "X Symmetry" ),
  _y_mirror( "Y Symmetry" ),
  _z_mirror( "Z Symmetry" ),
  _color( 88 ),

  _smoothed( "Smoothed" ),
  _double_sided( "Double Sided" ),
  _settings( "Settings" ),
  /*_bg_color( 0, 64, 168, 20 ),
  _history_size( 0, 84, 168, 20 ),*/
  
  _about( "About" ),
  _model_info( "Model Info" ),
  _status_bar( "No model loaded - click 'New' or 'Open' to begin" )
{
	ResizeSignal().connect( sigc::mem_fun( _v, &ViewGL::Refresh ) );
	_mesh_history.Changed().connect( sigc::mem_fun( _v, &ViewGL::Refresh ) );
	//ResizeSignal.connect( sigc::mem_fun( *this, &MainWindow::_resize ) );

	_brush.SetPositionEdit( true );
	_load_preferences();
	_initialize_widgets();
	
	Mesh::MainBrush = &_brush;

	Resize();
}

MainWindow::~MainWindow()
{
	_save_preferences();
}

void MainWindow::_draw()
{
	//TopWindow::_draw();
	
	glEnable( GL_ALPHA_TEST );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	glClearColor( 0.6, 0.6, 0.6, 1 );
	glClear( GL_COLOR_BUFFER_BIT );
	
	Reorganize();

	DrawChildren();
}

void MainWindow::Reorganize()
{
	glViewport( 0, 0, W(), H() );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( 0, Right(), Bottom(), 0 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( -0.5, -0.5, 0 ); // Ah, that helps!
}

void MainWindow::Refresh()
{	
	_save.SetEnabled( !_mesh_history.GetCurrentMesh().Saved() );
	_save_as.SetEnabled( _v.Visible() );
	_undo.SetEnabled( _mesh_history.UndoPossible() );
	_redo.SetEnabled( _mesh_history.RedoPossible() );
	_deform.SetEnabled( _v.Visible() );
	_model_info.SetEnabled( _v.Visible() );
}

void MainWindow::RefreshAll()
{
	Refresh();
	_cb_update_brush_settings();
}

void MainWindow::Resize()
{
	TopWindow::Resize();

	_v.SetX( 4 );
	_v.SetY( 92 );
	_v.SetW( W() - 8 );
	_v.SetH( H() - 116 );
	
	_bottom_bar.SetY( H() - 20 );
	_bottom_bar.SetW( W() );
	_status_bar.AnchoredSetRight( Right() );
}

bool MainWindow::IsMeshDataSafe()
{
	std::string response;
	const bool SAFE = true;
	const bool UNSAFE = false;
	
	if( _mesh_history.GetCurrentMesh().Saved() )
		return SAFE;
	else
	{		
		response = MessageDialog::ModalMessage( *this, "Unsaved data",
			"The model is not saved.\nSave before continuing?",
			"Don't Save", "Save", "Cancel" );
		
		if( response == "Cancel" )
		{
			_status_bar.SetLabel( "Operation canceled" );
			return UNSAFE;
		}
		else if( response == "Save" )
		{
			_save_model();
			if( _mesh_history.GetCurrentMesh().Saved() )
				return SAFE;
			else
				return UNSAFE;				
		}
		else if( response == "Don't Save" )
			return SAFE;
		else
			return UNSAFE;
	}
}

MeshHistory& MainWindow::GetMeshHistory()
{
	return _mesh_history;
}

void MainWindow::SetModelPath( std::string p )
{
	_model_path = p;
}
ViewGL& MainWindow::View()
{
	return _v;
}
void MainWindow::SetStatusText( std::string t )
{
	_status_bar.SetLabel( t );
	Draw();
}
void MainWindow::DisplayModelInfo()
{
	_status_bar.SetLabel( _mesh_history.GetCurrentMesh().GetModelStatisticsString() );
}

void MainWindow::_initialize_view()
{
	_v.SetVisible( false );
	_v._history = &_mesh_history;
	_v.MainBrush = &_brush;
}

bool MainWindow::_open_model( std::string path )
{
	if( IsMeshDataSafe() )
	{
		SetModelPath( path );
		GetMeshHistory().Clear();
		GetMeshHistory().GetCurrentMesh().Load( path );
		View().Show();
		View().ResetView();
		RefreshAll();
		SetStatusText( FileDialog::FilePart( path ) + " loaded, " +
			GetMeshHistory().GetCurrentMesh().GetModelStatisticsString() );
		return true;
	}
	else
		_status_bar.SetLabel( "Open canceled" );
	
	return false;
}

void MainWindow::_save_model()
{
	if( _mesh_history.GetCurrentMesh().FileName().length() )
		_save_model_as( _mesh_history.GetCurrentMesh().FileName() );
	else
		_save_dialog.Show();
}

void MainWindow::_save_model_as( std::string path )
{
	if( path.length() )
	{
		_mesh_history.GetCurrentMesh().Save( path );
		_model_path = path;
		Refresh();
	}
}

void MainWindow::_initialize_widgets()
{
	_deform_dialog.SetVisible( false );
	_open_dialog.OpenSignal().connect( sigc::mem_fun( *this, &MainWindow::_open_model ) );
	_new_dialog.SetVisible( false );
	_open_dialog.SetVisible( false );
	_open_dialog.ChangeDir( FileDialog::PathPart( _model_path ) );
	_save_dialog.SaveSignal().connect( sigc::mem_fun( *this, &MainWindow::_save_model_as ) );
	_save_dialog.SetVisible( false );
	_settings_dialog.SetVisible( false );
	_dialogs.PushBack( _deform_dialog );
	_dialogs.PushBack( _new_dialog );
	_dialogs.PushBack( _open_dialog );
	_dialogs.PushBack( _save_dialog );
	_dialogs.PushBack( _settings_dialog );
	_v.ChangedMeshSignal.connect( sigc::mem_fun( *this, &MainWindow::Refresh ) );
	
	_mesh_history.Changed().connect( sigc::mem_fun( *this, &MainWindow::Refresh ) );
	//x
	
	_model_left.PushBack( _load );
	_model_right.PushBack( _new );
	_model_left.PushBack( _save );
	_model_right.PushBack( _save_as );
	_model_left.PushBack( _undo );
	_model_right.PushBack( _redo );
	_model.PushBack( _model_left );
	_model.PushBack( _model_right );
	
	_brush_settings_mode.PushBack( _displace );
	_brush_settings_mode.PushBack( _grab );
	_brush_settings_mode.PushBack( _smooth );
	_brush_settings_mode.PushBack( _deform );
	_brush_settings_sliders.PushBack( _radius );
	_brush_settings_sliders.PushBack( _strength );
	_brush_settings_sliders.PushBack( _intensity );
	_brush_settings_push.PushBack( _add );
	_add.SetW( 67 );
	_brush_settings_push.PushBack( _sub );
	_brush_settings_push.PushBack( _rgb );
	_brush_settings_sliders.PushBack( _brush_settings_push );
	_brush_settings_symmetry.PushBack( _x_mirror );
	_brush_settings_symmetry.PushBack( _y_mirror );
	_brush_settings_symmetry.PushBack( _z_mirror );
	_brush_settings.PushBack( _brush_settings_mode );
	_brush_settings.PushBack( _brush_settings_sliders );
	_brush_settings.PushBack( _brush_settings_symmetry );
	
	_view.PushBack( _smoothed );
	_view.PushBack( _double_sided );
	_view.PushBack( _settings );
	
	_top_bar.PushBack( _model );
	_top_bar.PushBack( _brush_settings );
	_top_bar.PushBack( _color );
	_top_bar.PushBack( _view );
	
	_bottom_bar.PushBack( _model_info );
	_bottom_bar.PushBack( _about );
	_bottom_bar.PushBack( _status_bar );
	
	PushBack( _v );
	PushBack( _top_bar );
	PushBack( _bottom_bar );
	PushBack( _dialogs );
	
	// Load button
	_load.MouseUpSignal.connect( sigc::mem_fun( _open_dialog, &OpenDialog::Show ) );
	_load.SetShortcut( KeyData( 'o', false, true ) );
	// New button
	_new.MouseUpSignal.connect( sigc::mem_fun( _new_dialog, &NewDialog::Show ) );
	_new.SetShortcut( KeyData( 'n', false, true ) );
	// Save button
	_save.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_save_model ) );
	_save.SetShortcut( KeyData( 's', false, true ) );
	_save.SetEnabled( false );
	// Save As button
	_save_as.MouseUpSignal.connect( sigc::mem_fun( _save_dialog, &SaveDialog::Show ) );
	_save_as.SetEnabled( false );
	// Undo button
	_undo.SetShortcut( KeyData( 'z', false, true ) );
	_undo.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_undo ) );
	_undo.SetEnabled( false );
	// Redo button
	_redo.SetShortcut( KeyData( 'z', true, true ) );
	_redo.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_redo ) );
	_redo.SetEnabled( false );
	
	_displace.SetValue( 1 );
	_displace.SetShortcut( KeyData( 'd' ) );
	_displace.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_draw ) );
	_grab.SetShortcut( KeyData( 'g' ) );
	_grab.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_grab ) );
	_smooth.SetShortcut( KeyData( 's' ) );
	_smooth.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_smooth ) );
	_deform.SetEnabled( false );
	_deform.MouseUpSignal.connect( sigc::mem_fun( _deform_dialog, &DeformDialog::Show ) );
	_deform.SetShortcut( KeyData( 'd', false, true ) );
	_radius.SetProperties( 1, 100, 1 );
	_radius.Value( 10 );
	_radius.ChangedSignal.connect( sigc::hide< -1 >( sigc::mem_fun( *this, &MainWindow::_cb_update_brush_settings ) ) );
	_strength.SetProperties( 0, 100, 1 );
	_strength.Value( 25 );
	_strength.ChangedSignal.connect( sigc::hide< -1 >( sigc::mem_fun( *this, &MainWindow::_cb_update_brush_settings ) ) );
	_intensity.SetProperties( 0, 100, 1 );
	_intensity.Value( 25 );
	_intensity.ChangedSignal.connect( sigc::hide< -1 >( sigc::mem_fun( *this, &MainWindow::_cb_update_brush_settings ) ) );
	_add.PressToggle();
	_add.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_add ) );
	_sub.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_sub ) );
	_rgb.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_rgb ) );
	_x_mirror.SetShortcut( KeyData( 'x' ) );
	_x_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_symmetry ) );
	_y_mirror.SetShortcut( KeyData( 'y' ) );
	_y_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_symmetry ) );
	_z_mirror.SetShortcut( KeyData( 'z' ) );
	_z_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_symmetry ) );
	_color.ChangedSignal.connect( sigc::mem_fun( _brush, &Brush::SetColor ) );
	
	_smoothed.MouseUpSignal.connect( sigc::mem_fun( _v, &ViewGL::ToggleSmoothed ) );
	_smoothed.PressToggle();
	_double_sided.MouseUpSignal.connect( sigc::mem_fun( _v, &ViewGL::ToggleDoubleSided ) );
	_double_sided.SetValue( true );
	_settings.MouseUpSignal.connect( sigc::mem_fun( _settings_dialog, &SettingsDialog::Show ) );
	
	_model_info.MouseUpSignal.connect( sigc::mem_fun( this, &MainWindow::DisplayModelInfo ) );
	_model_info.SetEnabled( false );
	_about.MouseUpSignal.connect( sigc::mem_fun( this, &MainWindow::_display_about ) );
	
	_initialize_view();
	_cb_update_brush_settings();
}

void MainWindow::_display_about()
{
	_status_bar.SetLabel( "SharpConstruct v0.07, Copyright Nicholas Bishop, Licensed under the GPL - http://www.sharp3d.sourceforge.net" );
}

bool MainWindow::Close()
{
	if( IsMeshDataSafe() )
	{
		_save_preferences();
		return true;
	}
	
	return false;
}

void MainWindow::_cb_undo()
{
	_mesh_history.Undo();
	Refresh();
}

void MainWindow::_cb_redo()
{
	_mesh_history.Redo();
	Refresh();
}

void MainWindow::_cb_draw()
{
	_brush.SetModelingMode( SharpConstruct::Brush::DISPLACE );
	_displace.SetValue( 1 );
	_grab.SetValue( 0 );
	_smooth.SetValue( 0 );
}

void MainWindow::_cb_grab()
{
	_brush.SetModelingMode( SharpConstruct::Brush::GRAB );
	_displace.SetValue( 0 );
	_grab.SetValue( 1 );
	_smooth.SetValue( 0 );
}

void MainWindow::_cb_smooth()
{
	_brush.SetModelingMode( SharpConstruct::Brush::SMOOTH );
	_displace.SetValue( 0 );
	_grab.SetValue( 0 );
	_smooth.SetValue( 1 );
}

void MainWindow::_cb_update_brush_settings()
{
	float n_strength = _strength.Value() / -1000;
	
	if( !_add.Value() && !_sub.Value() )
		_brush.SetPositionEdit( false );
	else if( _sub.Value() )
	{
		n_strength = -n_strength;
		_brush.SetPositionEdit( true );
	}
	else if( _add.Value() )
	{
		_brush.SetPositionEdit( true );
	}
	
	_brush.SetRadius( _radius.Value() );
	_brush.AdjustRadius( _mesh_history.GetCurrentMesh().GetDiameter() );
	_brush.SetStrength( n_strength );
	_brush.SetIntensity( _intensity.Value() / 100 );
}

void MainWindow::_cb_add()
{
	if( _add.Value() && _sub.Value() )
		_sub.PressToggle();
	
	_cb_update_brush_settings();
}

void MainWindow::_cb_sub()
{
	if( _add.Value() && _sub.Value() )
		_add.PressToggle();
	
	_cb_update_brush_settings();
}

void MainWindow::_cb_rgb()
{
	_brush.SetColorEdit( _rgb.Value() );
}

void MainWindow::_cb_symmetry()
{
	_brush.SetSymmetryX( _x_mirror.Value() );
	_brush.SetSymmetryY( _y_mirror.Value() );
	_brush.SetSymmetryZ( _z_mirror.Value() );
}

void MainWindow::_load_preferences()
{	
	SetW( Preferences::Instance().Get( "W", 1024 ) );
	SetH( Preferences::Instance().Get( "H", 768 ) );
	_model_path = Preferences::Instance().Get( "MODEL_PATH",
	                                           FileDialog::HomePath() );
	_open_dialog.ChangeDir( FileDialog::PathPart( _model_path ) );
	_save_dialog.ChangeDir( FileDialog::PathPart( _model_path ) );

	_v.SetBackgroundColor( Color(
		Preferences::Instance().Get( "VIEW_COLOR_R", (float)0.3f ),
		Preferences::Instance().Get( "VIEW_COLOR_G", (float)0.3f ),
		Preferences::Instance().Get( "VIEW_COLOR_B", (float)0.3f ) ) );
	
	_mesh_history.SetMaximum( Preferences::Instance().Get( "HISTORY_STEPS", 2 ) );
}

void MainWindow::_save_preferences()
{
	Preferences::Instance().Set( "W", W() );
	Preferences::Instance().Set( "H", H() );
	Preferences::Instance().Set( "MODEL_PATH", _model_path );
}
