/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Mesh.h"

using SharpConstruct::Optimized::Point3D;

void Mesh::Move( float x, float y, float z )
{	
	for( int i = 0; i < _vertex_locations.size(); ++i )
	{
		_vertex_locations[ i ].X() += x;
		_vertex_locations[ i ].Y() += y;
		_vertex_locations[ i ].Z() += z;
	}
	
	_change_operation();
}

void Mesh::Resize( float x, float y, float z )
{
	for( int i = 0; i < _vertex_locations.size(); ++i )
	{
		if( x > 0 )
			_vertex_locations[ i ].X() *= x;
		else if( x < 0 )
			_vertex_locations[ i ].X() /= -x;
		if( y > 0 )
			_vertex_locations[ i ].Y() *= y;
		else if( y < 0 )
			_vertex_locations[ i ].Y() /= -y;
		if( z > 0 )
			_vertex_locations[ i ].Z() *= z;
		else if( z < 0 )
			_vertex_locations[ i ].Z() /= -z;
	}
	
	_recalculate_all_normals();
	_change_operation();
}

/** Smooths the entire mesh.
 *
 * Smooths every point in the mesh. Accepts a strength value between zero and
 * one.
**/
void Mesh::Smooth( float strength )
{
	unsigned j;
	int i, k, polygon_index, total, previous_point_index, next_point_index;
	Point3D average;
	
	for( i = 0; i < _vertex_locations.size(); ++i )
	{
		total = 0;
		average.Zero();
		for( j = 0; j < _vertex_polygon_indices[ i ].size(); j++ )
		{
			polygon_index = _vertex_polygon_indices[ i ][ j ];
			for( k = 0; k < 3; k++ )
				if( _indices[ polygon_index * 3 + k ] == i )
				{
					previous_point_index = k - 1;
					next_point_index = k + 1;
					if( previous_point_index < 0 )
						previous_point_index = 2;
					if( next_point_index == 3 )
						next_point_index = 0;
					
					average += _vertex_locations[ i ];
					average += _vertex_locations[ _indices[ polygon_index * 3 + previous_point_index ] ];
					average += _vertex_locations[ _indices[ polygon_index * 3 + next_point_index ] ];
					total += 3;
				}
		}
		average /= total;
	
		_vertex_locations[ i ] += ( average - _vertex_locations[ i ] ) * strength;
	}
	
	_recalculate_all_normals();
	_change_operation();
}

/** Floods the entire mesh with the current color.
 *
 * All vertex colors are shifted towards the current color. accepts an intensity
 * value between 0 and 1.
**/
void Mesh::Flood( float intensity )
{
	if( intensity == 1 )
	{
		for( unsigned i = 0; i < _vertex_colors.size(); i++ )
			_vertex_colors[ i ] = MainBrush->GetColor();
	}
	else
	{
		for( unsigned i = 0; i < _vertex_colors.size(); i++ )
		{
			_vertex_colors[ i ].Red = _vertex_colors[ i ].Red +
								   ( MainBrush->GetRed() - _vertex_colors[ i ].Red )
								   * intensity;
			_vertex_colors[ i ].Green = _vertex_colors[ i ].Green +
								   ( MainBrush->GetGreen() - _vertex_colors[ i ].Green )
								   * intensity;
			_vertex_colors[ i ].Blue = _vertex_colors[ i ].Blue +
								   ( MainBrush->GetBlue() - _vertex_colors[ i ].Blue )
								   * intensity;
		}
	}
	_change_operation();
}

void Mesh::Subdivide()
{
	unsigned max;
	GtsSurface *surface = gts_surface_new( gts_surface_class(), gts_face_class(),
	                                       gts_edge_class(), gts_color_vertex_class() );

	_to_gts_surface( *surface );
	
	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_indices.clear();
	_vertex_polygon_indices.clear();
	_polygon_normals.clear();
	
	/*FILE *fx = fopen( "/home/muvlo/x_wing.gts", "r" );
	GtsFile *gfile = gts_file_new( fx );
	if (gts_surface_read (surface, gfile))
	{
		fputs ("optimize: file on standard input is not a valid GTS file\n", stderr);
		fprintf (stderr, "stdin:%d:%d: %s\n", gfile->line, gfile->pos, gfile->error);
	}*/
	
	max = gts_surface_edge_number( surface ) * 2;
	gts_surface_refine( surface, 0, 0, ( GtsRefineFunc )_gts_segment_midvertex,
	                    0, _stop_gts_subdivide, &max );

	_from_gts_surface( *surface );
	_vertex_map.clear();
	gts_object_destroy( ( GtsObject * )( surface ) );
	_change_operation();
}

void Mesh::Optimize()
{
	unsigned min;
	GtsSurface *surface = gts_surface_new( gts_surface_class(), gts_face_class(),
	                                       gts_edge_class(), gts_color_vertex_class() );

	_to_gts_surface( *surface );
	
	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_indices.clear();
	_vertex_polygon_indices.clear();
	_polygon_normals.clear();

	min = gts_surface_edge_number( surface ) / 2;
	gts_surface_coarsen( surface, 0, 0, ( GtsRefineFunc )_gts_segment_midvertex,
	                     0, _stop_gts_coarsen, &min, 1 );
	
	_from_gts_surface( *surface );
	_vertex_map.clear();
	gts_object_destroy( ( GtsObject * )( surface ) );
	_change_operation();
}

int Mesh::RemoveDuplicates()
{
	return 0;
	/*unsigned int i, j, k, l, total = 0;
	
	for( i = 0; i < _vertices.size() - 1; i++ )
		for( j = i + 1; j < _vertices.size(); j++ )
			if( _vertices[ i ].Location == _vertices[ j ].Location )
			{
				// We've found a duplicate, destroy it
				_vertices.erase( _vertices.begin() + j );
				
				for( k = 0; k < _vertex_polygon_indices[ j ].size(); k++ )
				{
					// Copy all of the jth's polygons to i's vector
					_vertex_polygon_indices[ i ].push_back( _vertex_polygon_indices[ j ][ k ] );
					for( l = 0; l < 3; l++ )
						if( _indices[ _vertex_polygon_indices[ j ][ k * 3 + l ] ] == j )
							// Set any index equaling j to i
							_indices[ _vertex_polygon_indices[ j ][ k * 3 + l ] ] == i;
				}
				
				// Destroy the extra vertex_polygon_index
				_vertex_polygon_indices.erase( _vertex_polygon_indices.begin() + j );
				
				for( k = 0; k < _indices.size(); k++ )
					if( _indices[ k ] > j )
						_indices[ k ]--;
				total++;
				j--;
			}
	
	_recalculate_all_normals();
			
	return total;*/
}

void Mesh::Mirror( bool x, bool y, bool z )
{
	for( int i = 0; i < _vertex_locations.size(); ++i )
	{
		if( x )
			_vertex_locations[ i ].X() = -_vertex_locations[ i ].X();
		if( y )
			_vertex_locations[ i ].Y() = -_vertex_locations[ i ].Y();
		if( z )
			_vertex_locations[ i ].Z() = -_vertex_locations[ i ].Z();
	}
	
	if( !( ( x && y && !z ) || ( x && z && !y ) || ( y && z && !x ) ) )
		_flip_normals();
	
	
	_recalculate_all_normals();
	_change_operation();
}

void Mesh::Unify()
{
	_calculate_bounding_sphere();

	for( int i = 0; i < _vertex_locations.size(); ++i )
	{
		_vertex_locations[ i ].X() -= _center.X();
		_vertex_locations[ i ].Y() -= _center.Y();
		_vertex_locations[ i ].Z() -= _center.Z();
	}
	_change_operation();
}
