/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Mesh.h"

using SharpConstruct::Optimized::Point3D;

void Mesh::_to_gts_surface( GtsSurface &surface )
{
	std::vector< GtsColorVertex * > vertices;
	GList *edges = 0; // Using a GList here for the gts_edges_merge later on
	std::vector< GtsFace * > faces;
	GtsEdge *e1, *e2, *e3;

	vertices.resize( _vertex_locations.size() );
	for( unsigned i = 0; i < vertices.size(); i++ )
	{
		vertices[ i ] = ( GtsColorVertex * )gts_vertex_new( gts_color_vertex_class(),
		                              _vertex_locations[ i ].X(),
		                              _vertex_locations[ i ].Y(),
		                              _vertex_locations[ i ].Z() );
		vertices[ i ]->c.r = _vertex_colors[ i ].Red;
		vertices[ i ]->c.g = _vertex_colors[ i ].Green;
		vertices[ i ]->c.b = _vertex_colors[ i ].Blue;
	}
	
	faces.resize( _polygon_normals.size() );

	for( unsigned i = 0; i < faces.size(); i++ )
	{
		e1 = gts_edge_new( gts_edge_class(),
		                   ( GtsVertex * )vertices[ _indices[ i * 3 ] ],
		                   ( GtsVertex * )vertices[ _indices[ i * 3 + 1 ] ] );
		e2 = gts_edge_new( gts_edge_class(),
		                   ( GtsVertex * )vertices[ _indices[ i * 3 + 1 ] ],
		                   ( GtsVertex * )vertices[ _indices[ i * 3 + 2 ] ] );
		e3 = gts_edge_new( gts_edge_class(),
		                   ( GtsVertex * )vertices[ _indices[ i * 3 ] ],
		                   ( GtsVertex * )vertices[ _indices[ i * 3 + 2 ] ] );
		
		edges = g_list_prepend( edges, e1 );
		edges = g_list_prepend( edges, e2 );
		edges = g_list_prepend( edges, e3 );
		
		faces[ i ] = gts_face_new( gts_face_class(), e1, e2, e3 );
		gts_surface_add_face( &surface, faces[ i ] );
	}
	edges = gts_edges_merge( edges );
}

void Mesh::_add_gts_vertex( GtsColorVertex *vertex, gpointer *data )
{
	typedef std::map< GtsVertex *, int > V;
	
	int *index = static_cast< int* >( data[ 0 ] );
	Mesh &m = *static_cast< Mesh * >( data[ 1 ] );

	m._vertex_locations.push_back( Point3D( vertex->parent.p.x, vertex->parent.p.y, vertex->parent.p.z) );
	m._vertex_normals.push_back( Point3D() );
	m._vertex_colors.push_back( SharpConstruct::Color() );
	( *( m._vertex_colors.end() - 1 ) ).Red = vertex->c.r;
	( *( m._vertex_colors.end() - 1 ) ).Green = vertex->c.g;
	( *( m._vertex_colors.end() - 1 ) ).Blue = vertex->c.b;
	
	m._vertex_map[ vertex ] = *index;
	
	( *index )++;
}

void Mesh::_from_gts_surface( GtsSurface &surface )
{
	gpointer data[ 2 ];
	guint index = 0;
	data[ 0 ] = &index;
	data[ 1 ] = this;

	gts_surface_foreach_vertex( &surface, ( GtsFunc )_add_gts_vertex, data );
	gts_surface_foreach_face( &surface, ( GtsFunc )_add_gts_triangle, data );

	_polygon_normals.resize( _indices.size() / 3 );
	_vertex_polygon_indices.resize( _vertex_locations.size() );
	for( unsigned i = 1; i < _indices.size(); i++ )
		_vertex_polygon_indices[ _indices[ i ] ].push_back( i / 3 );
	
	_recalculate_all_normals();
}

void Mesh::_add_gts_triangle( GtsTriangle *triangle, gpointer *data )
{
	Mesh &m = *static_cast< Mesh * >( data[ 1 ] );
	
	int i1 = m._vertex_map[ ( GtsColorVertex * )triangle->e1->segment.v1 ],
	         i2 = m._vertex_map[ ( GtsColorVertex * )triangle->e1->segment.v2 ],
	         i3 = m._vertex_map[ ( GtsColorVertex * )triangle->e2->segment.v1 ],
	         i4 = m._vertex_map[ ( GtsColorVertex * )triangle->e2->segment.v2 ],
	         i5 = m._vertex_map[ ( GtsColorVertex * )triangle->e3->segment.v1 ],
	         i6 = m._vertex_map[ ( GtsColorVertex * )triangle->e3->segment.v2 ];

	m._indices.push_back( i1 );
	// I'm sure there's a nicer way to do this, but I don't know how
	if( i1 == i3 && i2 == i5 )
	{
		m._indices.push_back( i4 );
		m._indices.push_back( i5 );
	}
	else if( i1 == i3 && i2 == i6 )
	{
		m._indices.push_back( i4 );
		m._indices.push_back( i6 );
	}
	else if( i1 == i4 && i2 == i5 )
	{
		m._indices.push_back( i3 );
		m._indices.push_back( i5 );
	}
	else if( i1 == i4 && i2 == i6 )
	{
		m._indices.push_back( i3 );
		m._indices.push_back( i6 );
	}
	else if( i1 == i5 && i2 == i3 )
	{
		m._indices.push_back( i3 );
		m._indices.push_back( i6 );
	}
	else if( i1 == i5 && i2 == i4 )
	{
		m._indices.push_back( i4 );
		m._indices.push_back( i6 );
	}
	else if( i1 == i6 && i2 == i3 )
	{
		m._indices.push_back( i3 );
		m._indices.push_back( i5 );
	}
	else if( i1 == i6 && i2 == i4 )
	{
		m._indices.push_back( i4 );
		m._indices.push_back( i5 );
	}
	// There shouldn't be any other possibilites... something is seriously wrong
	// if there are.
}

gboolean Mesh::_stop_gts_coarsen( gdouble cost, guint number, gpointer min )
{
	if( number < *( int* )min )
		return true;
	return false;
}

gboolean Mesh::_stop_gts_subdivide( gdouble cost, guint number, gpointer max )
{
	if( number > *( int* )max )
		return true;
	return false;
}

GtsVertex * Mesh::_gts_segment_midvertex( GtsSegment *segment,
                                          GtsVertexClass *vertex_class )
{
	GtsColorVertex *vertex_1 = ( GtsColorVertex * )segment->v1,
	               *vertex_2 = ( GtsColorVertex * )segment->v2,
	               *vertex = ( GtsColorVertex * )gts_vertex_new( vertex_class,
	               // Sets the midpoint location
	               ( vertex_1->parent.p.x + vertex_2->parent.p.x ) / 2, 
	               ( vertex_1->parent.p.y + vertex_2->parent.p.y ) / 2,
	               ( vertex_1->parent.p.z + vertex_2->parent.p.z ) / 2);

	// Sets the "midcolor"
	vertex->c.r = ( vertex_1->c.r + vertex_2->c.r ) / 2;
	vertex->c.g = ( vertex_1->c.g + vertex_2->c.g ) / 2;
	vertex->c.b = ( vertex_1->c.b + vertex_2->c.b ) / 2;
	
	return ( GtsVertex * )vertex;
}
