/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Mesh.h"
#include "Polygon3D.h"
#include "Utilities.h"
#include <cmath>
#include <fstream>
#include <sstream>
#include <string>

#include <iostream>

using SharpConstruct::Optimized::Normal3D;
using SharpConstruct::Optimized::Point3D;

Normal3D Mesh::_view_up_normal( 0, 0, 0 );
Normal3D Mesh::_view_right_normal( 0, 0, 0 );
SharpConstruct::Brush *Mesh::MainBrush = 0;
Point3D Mesh::_current_selection( 0, 0, 0 );
//Magick::Image *Mesh::_texture = NULL;

Mesh::Mesh()
: _saved( true )
{}

/** Loads any supported mesh type.
 *
 * Parses the filename string to find the extension, then calls the appropriate
 * function to do the actual loading.
**/
void Mesh::Load( std::string filename )
{
	std::string extension = filename.substr( filename.rfind( '.',
	                                         std::string::npos ) + 1 );
	
	transform( extension.begin(), extension.end(), extension.begin(), toupper );

	if( extension == "OBJ" )
		_load_obj( filename );
	
	_filename = filename;
	_saved = true;
}

/** Saves any supported mesh type.
 *
 * Parses the filename string to find the extension, then calls the appropriate
 * function to do the actual saving.
**/
void Mesh::Save( std::string filename )
{
	std::string extension = filename.substr( filename.rfind( '.',
	                                         std::string::npos ) + 1 );
	
	transform( extension.begin(), extension.end(), extension.begin(), toupper );

	if( extension == "OBJ" )
		_save_obj( filename );
	else if( extension == "WRL" )
		_save_wrl( filename );
	
	_filename = filename;
	_saved = true;
}

void Mesh::BecomePlane( int columns, int rows )
{
	_filename = "";
	
	int i, j, t;
	const float half_x = columns / 2, half_y = rows / 2;
	
	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_indices.clear();
	_polygon_normals.clear();
	_vertex_polygon_indices.clear();
	
	//_vertex_locations.resize( rows * columns );
	_vertex_normals.resize( rows * columns );
	_vertex_colors.resize( rows * columns );
	for( i = 0; i < rows; i++ )
	{
		for( j = 0; j < columns; j++ )
		{
			_vertex_locations.push_back( Point3D( ( j - half_x ) / ( columns - 1 ), ( i - half_y ) / ( rows - 1 ), 0 ) );

			//_vertices[ _vertices.size() - 1 ].U = 1.0 * j / rows;
			//_vertices[ _vertices.size() - 1 ].V = 1 - ( 1.0 * i / rows );
		}
	}
	_vertex_polygon_indices.resize( _vertex_locations.size() );
	_polygon_normals.resize( ( rows - 1 ) * ( columns - 1 ) * 2 );
	for( i = 0, t = 0; i < rows - 1; i++ )
		for( j = 0; j < columns - 1; j++, t++ )
		{
			// Create two triangles
			_indices.push_back( i * columns + j + 1 );
			_indices.push_back( ( i + 1 ) * columns + j );
			_indices.push_back( i * columns + j );
			_indices.push_back( i * columns + j + 1 );
			_indices.push_back( ( i + 1 ) * columns + j + 1 );
			_indices.push_back( ( i + 1 ) * columns + j );
			
			_vertex_polygon_indices[ i * columns + j + 1 ].push_back( t );
			_vertex_polygon_indices[ ( i + 1 ) * columns + j ].push_back( t );
			_vertex_polygon_indices[ i * columns + j ].push_back( t );
			t++;
			_vertex_polygon_indices[ i * columns + j + 1 ].push_back( t );
			_vertex_polygon_indices[ ( i + 1 ) * columns + j + 1 ].push_back( t );
			_vertex_polygon_indices[ ( i + 1 ) * columns + j ].push_back( t );
			
			
			/*std::cout << i * columns + j + 1 << " "
			          << ( i + 1 ) * columns + j << " "
			          << i * columns + j << " "
			          << i * columns + j + 1 << " "
			          << ( i + 1 ) * columns + j + 1 << " "
			          << ( i + 1 ) * columns + j << std::endl;*/
		}


		
	_recalculate_all_normals();

	Unify();
	Flood( 1 );

	/*int i, j, t;
	float half_x = columns / 2, half_y = rows / 2;
	
	_vertices.Clear();
	_vertices.Resize( rows * columns );
	_indices.clear();
	_vertex_polygon_indices.clear();
	
	for( i = 0; i < rows; i++ )
		for( j = 0; j < columns; j++ )
		{
			_vertex_locations.Set( i * rows + j,
			                         ( j - half_x ) / ( columns - 1 ),
			                         ( i - half_y ) / ( rows - 1 ), 0 );
			//_vertices.U[ _vertices.Size() - 1 ] = 1.0 * j / rows;
			//_vertices.V[ _vertices.Size() - 1 ] = 1 - ( 1.0 * i / rows );
		}
	
	_vertex_polygon_indices.resize( _vertex_locations.size() );
	_polygon_normals.Clear();
	_polygon_normals.Resize( ( rows - 1 ) * ( columns - 1 ) * 2 );
	for( i = 0, t = 0; i < rows - 1; i++ )
		for( j = 0; j < columns - 1; j++, t += 2 )
		{
			// Create two triangles
			_indices.push_back( i * columns + j + 1 );
			_indices.push_back( ( i + 1 ) * columns + j );
			_indices.push_back( i * columns + j );
			_indices.push_back( i * columns + j + 1 );
			_indices.push_back( ( i + 1 ) * columns + j + 1 );
			_indices.push_back( ( i + 1 ) * columns + j );
			
			_vertex_polygon_indices[ i * columns + j + 1 ].push_back( t );
			_vertex_polygon_indices[ ( i + 1 ) * columns + j ].push_back( t );
			_vertex_polygon_indices[ i * columns + j ].push_back( t );
			_vertex_polygon_indices[ i * columns + j + 1 ].push_back( t + 1 );
			_vertex_polygon_indices[ ( i + 1 ) * columns + j + 1 ].push_back( t + 1 );
			_vertex_polygon_indices[ ( i + 1 ) * columns + j ].push_back( t + 1 );
		}

	*/
	/*struct timeval tp1, tp2;
	gettimeofday( &tp1, 0 );
	for( i = 0; i < 30000; i++ )
		_recalculate_all_normals();
	gettimeofday( &tp2, 0 );
	std::cout << ( tp2.tv_sec + 1.e-6 * tp2.tv_usec ) - 
	             ( tp1.tv_sec + 1.e-6 * tp1.tv_usec ) << std::endl;*/
	/*_recalculate_all_normals();

	Unify();
	Flood( 1 );*/
	
	_change_operation();
}

void Mesh::BecomeSphere( int subdivisions )
{
	_filename = "";
	
	GtsSurface *surface = gts_surface_new( gts_surface_class(), gts_face_class(),
	                                       gts_edge_class(), gts_vertex_class() );

	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_indices.clear();
	_vertex_polygon_indices.clear();
	_polygon_normals.clear();

	gts_surface_generate_sphere( surface, subdivisions );
	
	_from_gts_surface( *surface );
	_vertex_map.clear();
	gts_object_destroy( ( GtsObject * )( surface ) );
	
	Flood( 1 );
	
	_change_operation();
}

/*void Mesh::SetTexture( Magick::Image *t )
{
	_texture = t;
}*/

sigc::signal< void >& Mesh::Changed()
{
	return _changed;
}
void Mesh::_change_operation()
{
	_saved = false;
	Changed()();
}
std::string Mesh::FileName() const
{
	return _filename;
}
bool Mesh::Saved() const
{
	return _saved;
}

/** Returns mesh data information.
 *
 * Returns a string in this format:
 * "[number of vertices] vert[ex/ices], [number of polygons] polygon[s]"
**/
std::string Mesh::GetModelStatisticsString()
{
	std::ostringstream oss;

	oss << _vertex_locations.size();
	if( _vertex_locations.size() == 1 )
		oss << " vertex, ";
	else
		oss << " vertices, ";
	oss << _polygon_normals.size();
	if( _polygon_normals.size() == 1 )
		oss << " polygon";
	else
		oss << " polygons";

	return oss.str();
}

float Mesh::GetDiameter()
{
	_calculate_bounding_sphere();

	return _diameter;
}

void Mesh::CopyEssentials( Mesh &mesh )
{
	_vertex_locations = mesh._vertex_locations;
	_vertex_normals = mesh._vertex_normals;
	_vertex_colors = mesh._vertex_colors;
	_polygon_normals = mesh._polygon_normals;
	_indices = mesh._indices;
	_vertex_polygon_indices = mesh._vertex_polygon_indices;
	_saved = mesh._saved;
	_filename = mesh._filename;
}

/** Free unimportant mesh data.
 *
 * Free memory by clearing temporary data unimportant to a mesh object.
**/
void Mesh::ClearExtraData()
{
	_active_vertex_indices.clear();
	_active_vertex_distances.clear();
	_modified_vertex_indices.clear();
}

/** Loads a mesh in the OBJ format.
 *
 * Loads mesh data from an OBJ file. Support is incomplete; the parsing has
 * bugs, and normal data is ignored.
**/
/*void Mesh::_load_obj( std::string filename )
{
	std::ifstream file( filename.c_str(), std::ios::in );
	std::string token, line_string, face_group;
	int vertex_index = 0, polygon_index = 0;
	unsigned find_loc;
	unsigned tmp;
	bool hascolor = false;
	//std::stringstream line;

	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_polygon_normals.clear();
	_vertex_polygon_indices.clear();
	_indices.clear();
	std::vector< Polygon3D > polygons;
	
	while( !file.eof() )
	{
		std::stringstream line;
		std::getline( file, line_string );
		line.str( line_string );
		token.clear();
		line >> token;
		if( token == "v" )
		{
			_vertex_locations.resize( vertex_index + 1 );
			line >> _vertex_locations[ vertex_index ].X() >>
			        _vertex_locations[ vertex_index ].Y() >>
			        _vertex_locations[ vertex_index ].Z();
			vertex_index++;
		}
		else if( token == "f" )
		{
			polygons.resize( polygon_index + 1 );

			while( line >> face_group )
			{
				int slash = face_group.find( '/' );
				polygons[ polygon_index ].VertexIndices.push_back(
					SharpConstruct::ToInt( face_group.substr( 0, slash ) ) );
				face_group.clear();
			}
			
			polygon_index++;
		}
	}
	
	///
	polygon_index = 0;
	_vertex_polygon_indices.resize( _vertex_locations.size() );
	_vertex_normals.resize( _vertex_locations.size() );
	
	for( unsigned i = 0; i < polygons.size(); i++ )
		for( unsigned j = 1; j < polygons[ i ].VertexIndices.size() - 1; j++ )
		{
			_indices.push_back( polygons[ i ].VertexIndices[ 0 ] );
			//_vertex_polygon_indices[ polygons[ i ].VertexIndices[ 0 ] ].push_back( polygon_index );
			_indices.push_back( polygons[ i ].VertexIndices[ j ] );
			//_vertex_polygon_indices[ polygons[ i ].VertexIndices[ j ] ].push_back( polygon_index );
			_indices.push_back( polygons[ i ].VertexIndices[ j + 1 ] );
			//_vertex_polygon_indices[ polygons[ i ].VertexIndices[ j + 1 ] ].push_back( polygon_index );
			polygon_index++;
		}
	
	_polygon_normals.resize( polygon_index );
	
	_recalculate_all_normals();
		
	_saved = true;*/
void Mesh::_load_obj( std::string filename )
{
	std::ifstream file( filename.c_str(), std::ios::in );
	std::string token;
	int vertex_count = 0, polygon_count = 0;
	unsigned find_loc;
	unsigned tmp;
	bool hascolor = false;

	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_polygon_normals.clear();
	_vertex_polygon_indices.clear();
	_indices.clear();
	std::vector< Polygon3D > polygons;
	//std::vector< float > u;
	//std::vector< float > v;

	while( !file.eof() )
	{
		file >> token;
		if( token == "v" )
		{
			_vertex_locations.resize( _vertex_locations.size() + 1 );
			file >> _vertex_locations[ vertex_count ].X() >>
			        _vertex_locations[ vertex_count ].Y() >>
			        _vertex_locations[ vertex_count ].Z();
			vertex_count++;
		}
		/*else if( token == "vt" )
		{
			u.resize( uv_count + 1 );
			v.resize( uv_count + 1 );
			file >> u[ uv_count ] >> v[ uv_count ];
			uv_count++;
		}*/
		else if( token == "f" )
		{
			polygons.resize( polygon_count + 1 );
			do
			{
				file >> token;
				if( token == "f" )
				{
					file.putback( 'f' );
					file.putback( ' ' );
					break;
				}
				find_loc = token.find( '/', 0 );
				if( find_loc == std::string::npos )
					polygons[ polygon_count ].VertexIndices.push_back( atoi( token.c_str() ) - 1 );
				else
					polygons[ polygon_count ].VertexIndices.push_back( atoi( token.substr( 0, find_loc ).c_str() ) - 1 );
				// This avoids the problem of the parsing not ending when
				// the last line has a space after the last number group
				if( file.peek() == ' ' )
					file.get();
			} while( file.peek() != '\n' && !file.eof() );
			polygon_count++;
		}
		else if( token == "#vertexcolor" )
		{
			file >> tmp;
			_vertex_colors.push_back( SharpConstruct::Color() );
			file >> _vertex_colors[ tmp ].Red >> _vertex_colors[ tmp ].Green >>
			        _vertex_colors[ tmp ].Blue;
			hascolor = true;
		}
	}
	file.close();
	
	/*if( u.size() && v.size() )
	{
		for( unsigned i = 0; i < u.size(); i++ )
		{
			_vertices.U[ i ] = u[ i ];
			_vertices.V[ i ] = v[ i ];
		}
	}*/
	
	polygon_count = 0;
	_vertex_polygon_indices.resize( _vertex_locations.size() );
	_vertex_normals.resize( _vertex_locations.size() );
	
	for( unsigned i = 0; i < polygons.size(); i++ )
		for( unsigned j = 1; j < polygons[ i ].VertexIndices.size() - 1; j++ )
		{
			_indices.push_back( polygons[ i ].VertexIndices[ 0 ] );
			_vertex_polygon_indices[ polygons[ i ].VertexIndices[ 0 ] ].push_back( polygon_count );
			_indices.push_back( polygons[ i ].VertexIndices[ j ] );
			_vertex_polygon_indices[ polygons[ i ].VertexIndices[ j ] ].push_back( polygon_count );
			_indices.push_back( polygons[ i ].VertexIndices[ j + 1 ] );
			_vertex_polygon_indices[ polygons[ i ].VertexIndices[ j + 1 ] ].push_back( polygon_count );
			polygon_count++;
		}
	
	_polygon_normals.resize( polygon_count );
	
	_recalculate_all_normals();
	
	if( !hascolor )
	{
		_vertex_colors.resize( _vertex_locations.size() );
		Flood( 1 );
	}
	
	_saved = true;
}

/** Saves a mesh in the OBJ format.
 *
 * Saves mesh data to an OBJ file. Only vertex and polygon data is exported,
 * not normals. The top line is a comment including the SharpConstruct URL.
**/
void Mesh::_save_obj( std::string filename )
{
	std::ofstream file( filename.c_str(), std::ios::out );
	
	int i, j;
	
	file << "# Exported from SharpConstruct 0.05 - http://sharp3d.sourceforge.net/\n\n";
	
	for( i = 0; i < _vertex_locations.size(); i++ )
		file << "v " << _vertex_locations[ i ].X() << ' ' <<
	                    _vertex_locations[ i ].Y() << ' ' <<
	                    _vertex_locations[ i ].Z() << '\n';
	
	//for( i = 0; i < _vertex_locations.size(); i++ )
	//	file << "vt " << _vertices.U[ i ] << ' ' << _vertices.V[ i ] << '\n';
	
	for( i = 0; i < _polygon_normals.size(); i++ )
	{
		file << "f ";
		for( j = 0; j < 3; j++ )
			file << _indices[ i * 3 + j ] + 1 << ' ';		
		file << '\n';
	}
	
	for( i = 0; i < _vertex_colors.size(); i++ )
		file << "#vertexcolor " << i << ' ' << _vertex_colors[ i ].Red << ' '
		     << _vertex_colors[ i ].Green << ' ' << _vertex_colors[ i ].Blue
		     << '\n';
	
	file.close();
	
	_saved = true;
}

/** Saves a mesh in the VRML97 (.wrl) format.
 *
 * Saves mesh data to an WRL file. Only vertex and polygon data is exported,
 * not normals. The top line is the VRML magic numbers string,
 * followed by a line with a comment including the SharpConstruct URL.
 *
 * Thanks to Joerg Scheurich for this code.
**/
void Mesh::_save_wrl( std::string filename )
{
	std::ofstream file( filename.c_str(), std::ios::out );
	
	int i, j;
	
	file << "#VRML V2.0 utf8\n# Exported from SharpConstruct 0.0? - http://sharp3d.sourceforge.net/\n\n";
	file << "Transform\n";
	file << "  {\n";
	file << "  children\n";
	file << "    [\n";
	file << "    Shape\n";
	file << "      {\n";
	file << "      appearance       Appearance\n";
	file << "        {\n";
	file << "        material        Material\n";
	file << "          {\n";
	file << "          }\n";
	file << "        }\n";
	file << "      geometry       IndexedFaceSet\n";
	file << "        {\n";
	file << "        # double sided\n";
	file << "        solid FALSE\n";
	file << "        # single sided: inside or outside ?\n";
	file << "        #ccw TRUE\n";
	file << "        #ccw FALSE\n";
	file << "        # smooth if angle < 90 degree\n";
	file << "        creaseAngle 1.57\n";
	file << "        coord   Coordinate\n";
	file << "          {\n";
	file << "          point\n";
	file << "            [\n";
	for( i = 0; i < _vertex_locations.size(); ++i )
	{
		file << "            " << 
		_vertex_locations[ i ].X() << ' ' <<
		_vertex_locations[ i ].Y() << ' ' <<
		_vertex_locations[ i ].Z() << '\n';
	}
	file << "            ]\n";
	file << "          }\n";
	file << "        coordIndex\n";
	file << "          [\n";
	for( i = 0; i < _polygon_normals.size(); ++i )
	{
		file << "          ";
		for( j = 0; j < 3; j++ )
			file << _indices[ i * 3 + j ] << ' ';		
		file << " -1 " << '\n';
	}
	file << "          ]\n";

/* not implemented yet	
for( i = 0; i < _vertices.size(); i++ )
	file << "#vertexcolor " << i << ' ' << _vertices[ i ].R << ' ' <<
			_vertices[ i ].G << ' ' << _vertices[ i ].B << '\n';
*/
	file << "        }\n";
	file << "      }\n";
	file << "    ]\n";
	file << "  }\n";

	file.close();
	
	_saved = true;
}

/** Calculates center and diameter of sphere formed by vertices.
 *
 * All vertices in the model are used to create a virtual sphere, with a center
 * at the average of all the points and the diameter equal to two times the
 * distance from the center to the furthest point from the center.
**/
void Mesh::_calculate_bounding_sphere()
{
	int i;
	float tmp;

	_diameter = 0;
	
	_center.Zero();

	for( i = 0; i < _vertex_locations.size(); i++ )
		_center += _vertex_locations[ i ];
	
	_center /= _vertex_locations.size();
	
	for( i = 0; i < _vertex_locations.size(); i++ )
	{
		tmp = _center.Distance( _vertex_locations[ i ] );
		if( tmp > _diameter )
			_diameter = tmp;
	}
	
	_diameter *= 2;
}

void Mesh::_flip_normals()
{
	std::vector< unsigned > indices;

	for( unsigned i = 0; i < _indices.size(); i += 3 )
		for( int j = 2; j >= 0; j-- )
			indices.push_back( _indices[ i + j ] );

	_indices = indices;
}

/** Calculates the average normal of the area.
 *
 * Returns the average normal of all the points within the area.
**/
Point3D Mesh::_calculate_area_normal()
{
	Normal3D area_normal;
	area_normal.Zero();
		
	for( unsigned i = 0; i < _active_vertex_indices.size(); i++ )
		area_normal += _vertex_normals[ _active_vertex_indices[ i ] ];
	
	area_normal.Normalize();
	
	return area_normal;
}

/** Updates all vertex and polygon normals.
 *
 * Acts like _recalculate_modified_normals(), but does the recalculation for
 * every vertex and every polygon.
**/
void Mesh::_recalculate_all_normals()
{
	for( int i = 0; i < _polygon_normals.size(); i++ )
		_polygon_normals[ i ].CalculatePlaneNormal(
			_vertex_locations[ _indices[ i * 3 ] ].RawData(),
			_vertex_locations[ _indices[ i * 3 + 1 ] ].RawData(),
			_vertex_locations[ _indices[ i * 3 + 2 ] ].RawData() );

	unsigned j;
	for( unsigned i = 0; i < _vertex_polygon_indices.size(); i++ )
	{
		_vertex_normals[ i ].Zero();
		for( j = 0; j < _vertex_polygon_indices[ i ].size(); j++ )
			_vertex_normals[ i ] += _polygon_normals[ _vertex_polygon_indices[ i ][ j ] ];

		_vertex_normals[ i ] /= j;
	}
}

/** Updates modified vertex and polygon normals.
 *
 * Uses the _modified_vertex_indices list to update only modified vertex and
 * polygon normals. This is generally much faster than calling
 * _recalculate_all_normals().
**/
void Mesh::_recalculate_modified_normals()
{
	unsigned i, j;
	
	//std::cout << "Polys: " << _polygon_normals.size();
	if( _modified_vertex_indices.size() < _vertex_locations.size() )
	{
		//std::vector< bool > used_poly( _polygon_normals.size(), false );
		for( i = 0; i < _modified_vertex_indices.size(); i++ )
		{
			for( j = 0; j < _vertex_polygon_indices[
				_modified_vertex_indices[ i ] ].size(); j++ )
			{
				unsigned poly = _vertex_polygon_indices[
					_modified_vertex_indices[ i ] ][ j ];
				/*if( used_poly[ poly ] == false )
				{*/
					_damaged_polygon_indices.push_back( poly );
					/*used_poly[ poly ] = true;
				}*/
			}
			/*_damaged_polygon_indices.insert( _damaged_polygon_indices.begin(),
				_vertex_polygon_indices[
				_modified_vertex_indices[ i ] ].begin(),
				_vertex_polygon_indices[
				_modified_vertex_indices[ i ] ].end() );*/
		}
		_damaged_polygon_indices.sort();
		_damaged_polygon_indices.unique();
		//std::cout << ", Damaged: " << _damaged_polygon_indices.size();
		
		/*for( i = 0; i < _modified_vertex_indices.size(); i++ )
			for( j = 0; j < _vertex_polygon_indices[ _modified_vertex_indices[ i ] ].size(); j++ )
			{
				const int index = _vertex_polygon_indices[ _modified_vertex_indices[ i ] ][ j ];
				_polygon_normals[ index ].CalculatePlaneNormal( 
					_vertex_locations[ _indices[ index * 3 ] ].RawData(),
					_vertex_locations[ _indices[ index * 3 + 1 ] ].RawData(),
					_vertex_locations[ _indices[ index * 3 + 2 ] ].RawData() );
			}*/
		for( std::list< unsigned >::iterator a = _damaged_polygon_indices.begin();
		     a != _damaged_polygon_indices.end(); a++ )
		{
			_polygon_normals[ *a ].CalculatePlaneNormal(
				_vertex_locations[ _indices[ ( *a ) * 3 ] ].RawData(),
				_vertex_locations[ _indices[ ( *a ) * 3 + 1 ] ].RawData(),
				_vertex_locations[ _indices[ ( *a ) * 3 + 2 ] ].RawData() );
		}
		for( i = 0; i < _modified_vertex_indices.size(); i++ )
		{
			_vertex_normals[ _modified_vertex_indices[ i ] ] = _polygon_normals[ _vertex_polygon_indices[ _modified_vertex_indices[ i ] ][ 0 ] ];
			for( j = 1; j < _vertex_polygon_indices[ _modified_vertex_indices[ i ] ].size(); j++ )
			{
				_vertex_normals[ _modified_vertex_indices[ i ] ]+= _polygon_normals[ _vertex_polygon_indices[ _modified_vertex_indices[ i ] ][ j ] ];
			}
			_vertex_normals[ _modified_vertex_indices[ i ] ] /= j;
		}
	}
	else
		_recalculate_all_normals();
}

/** Stores vertices within the MainBrush->GetRadius().
 *
 * Stores a list of all vertices within the sphere that has its center at area
 * and a radius of MainBrush->GetRadius(). The distance from the active vertices
 * to the center is also stored.
**/
// Slow
/*void Mesh::_find_active_vertices( std::vector< int > &vertex_indices,
                                  std::vector< float > &vertex_distances,
                                  Point3D area, int mirror_index )
{
	vertex_indices.clear();
	vertex_distances.clear();
	
	float distance;
	
	for( int i = 0; i < _vertex_locations.size(); i++ )
	{
		//distance = sqrt( pow( area.X() -
	    //                      _vertices[ i ].Location.X(), 2 ) +
	    //                 pow( area.Y() -
	    //                      _vertices[ i ].Location.Y(), 2 ) +
	    //                 pow( area.Z() -
	    //                      _vertices[ i ].Location.Z(), 2 ) );

		distance = area.Distance( _vertex_locations[ i ] );
		if( distance < MainBrush->GetRadius() )
		{
			if( i < _lowest_index )
				_lowest_index = i;
			if( i > _highest_index )
				_highest_index = i;
			vertex_indices.push_back( i );
			vertex_distances.push_back( distance );
		}
	}
	
	_selected_vertex[ mirror_index ] = vertex_indices[ 0 ];

	_modified_vertex_indices.insert( _modified_vertex_indices.end(),
		                             vertex_indices.begin(),
	                                 vertex_indices.end() );
}*/
/* This version of _find_active_vertices may be rather slow too...*/
void Mesh::_find_active_vertices( std::vector< int > &vertex_indices,
                                  std::vector< float > &vertex_distances,
                                  Point3D area, int mirror_index )
{	
	vertex_indices.clear();
	vertex_distances.clear();
	
	int area_index = 0;

	std::set< int > checked;
	
	if( _selected_vertex[ mirror_index ] >= _vertex_locations.size() )
		_selected_vertex[ mirror_index ] = 0;

	if( _vertex_locations[ _selected_vertex[ mirror_index ] ].Distance( area ) <
	    MainBrush->GetRadius() )
	{
		area_index = _selected_vertex[ mirror_index ];
		//std::cout << _vertex_locations[ _selected_vertex[ mirror_index ] ].Distance( area ) << std::endl;
	}
	else
	{
		float tmp_distance, smallest_distance = 99999; // Evil, I know...
		
		for( int i = 0; i < _vertex_locations.size(); i++ )
		{
			tmp_distance = fabs( area.X() - _vertex_locations[ i ].X() ) +
						   fabs( area.Y() - _vertex_locations[ i ].Y() ) +
						   fabs( area.Z() - _vertex_locations[ i ].Z() );
			if( tmp_distance < smallest_distance )
			{
				area_index = i;
				smallest_distance = tmp_distance;
				if( smallest_distance < 0.01 ) // (More than) close enough...
					break;
			}
		}
		
		//std::cout << smallest_distance << std::endl;
	}

	_is_vertex_active( area_index, checked, vertex_indices, vertex_distances,
	                   area );
	
	_selected_vertex[ mirror_index ] = vertex_indices[ 0 ];

	_modified_vertex_indices.insert( _modified_vertex_indices.end(),
		                             vertex_indices.begin(),
	                                 vertex_indices.end() );
}

void Mesh::_is_vertex_active( unsigned current, std::set< int > &checked,
                              std::vector< int > &vertex_indices,
                              std::vector< float > &vertex_distances,
                              Point3D area )
{
	checked.insert( current );
	
	float distance = area.Distance( _vertex_locations[ current ] );
	if( distance < MainBrush->GetRadius() )
	{
		if( current < _lowest_index )
			_lowest_index = current;
		if( current > _highest_index )
			_highest_index = current;
		vertex_indices.push_back( current );
		vertex_distances.push_back( distance );
	}
	else
		return;
	
	int j, polygon_index;
	int previous_point_index;
	// Loop through each polygon using the current point
	for( unsigned i = 0; i < _vertex_polygon_indices[ current ].size(); i++ )
	{
		// Set the current polygon
		polygon_index =_vertex_polygon_indices[ current ][ i ];
		// Loop through each vertex in the polygon
		for( j = 0; j <= 2; j++ )
			// If the current active point equals the current VertexIndex...
			if( _indices[ polygon_index * 3 + j ] == current )
			{
				previous_point_index = j - 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;

				previous_point_index = _indices[ polygon_index * 3 + previous_point_index ];

				if( checked.find( previous_point_index ) == checked.end() )
					_is_vertex_active( previous_point_index, checked,
					                   vertex_indices, vertex_distances,
					                   area );
				break;
			}
	}
}
