/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Slider.h"
#include "Text.h"
#include "Utilities.h"
#include <GL/gl.h>
#include <sstream>

#include <iostream>

using namespace SharpConstruct::Interface;

void Slider::KeyDown()
{
	if( Focused() )
	{
		if( Event::Keyboard.LeftArrow )
		{
			_sel_start--;
			_sel_all = false;
		}
		else if( Event::Keyboard.RightArrow )
		{
			_sel_start++;
			_sel_all = false;
		}
		else if( Event::Keyboard.Backspace )
		{
			std::string v;
			v = ToString( Value() );
			if( _sel_all )
			{
				Value( _min );
				v = ToString( Value() );
				_sel_start = v.length();
				_sel_all = false;
			}
			else if( _sel_start != 0 )
			{
				v.erase( _sel_start - 1, 1 );
				Value( ToFloat( v ) );
				_sel_start--;
				_sel_all = false;
			}
		}
		else if( Event::Keyboard.Enter )
		{
			SetSignal();
			if( !Steady() )
				Value( Anchor() );
		}
		else if( Event::Keyboard.Key >= '0' && Event::Keyboard.Key <= '9' )
		{
			std::string v;
			v = ToString( Value() );
			
			int len = 0;
			if( _sel_all )
				len = v.length();
			v.replace( _sel_start, len, ToString( Event::Keyboard.Key ) );
			Value( ToFloat( v ) );
			_sel_start++;
			_sel_all = false;
		}
		if( _sel_start < 0 )
			_sel_start = 0;
		else if( _sel_start > ToString( Value() ).length() )
			_sel_start = ToString( Value() ).length();
		ChangedSignal( Value() );
	}
}

void Slider::MakeInactive()
{
	Widget::MakeInactive();
	if( !Contains() )
		MouseUpSignal.emit();
}

void Slider::Value( float v )
{
	_value = round( ( v - _min ) / _step ) * _step + _min;

	if( _value < _min )
		_value = _min;
	else if( _value > _max )
		_value = _max;
	
	_orig = Value();
}

void Slider::SetProperties( float min, float max, float step, float anchor,
                            bool steady )
{
	SetProperties( min, max, step );
	SetAnchor( anchor );
	SetSteady( steady );
}

float Slider::Anchor() const
{
	return _anchor;
}
void Slider::SetAnchor( float a )
{
	if( a < _min )
		_anchor = _min;
	else if( a > _max )
		_anchor = _max;
	else
		_anchor = a;
}
bool Slider::Steady() const
{
	return _steady;
}
void Slider::SetSteady( bool s )
{
	_steady = s;
}

float Slider::LocationOf( float v ) const
{
	return X() + ( ( W() - 1 ) * ( v - _min ) ) / Range();
}

void Slider::_draw()
{
	// Draw border
	if( Contains() && Enabled() )
		glColor4f( 0.3, 0.3, 0.3, 1 );
	else
		glColor4f( 0.5, 0.5, 0.5, 1 );
	glRecti( X(), Y(), Right() + 1, Bottom() + 1 );
	
	 // Fill inside
	if( !Contains() || !Enabled() )
		glColor4f( 0.7, 0.7, 0.7, 1 );
	else if( Contains( Event::Mouse.DownLocation ) && Active() )
		glColor4f( 0.85, 0.85, 0.85, 1 );
	else
		glColor4f( 0.8, 0.8, 0.8, 1 );
	glRecti( X() + 1, Y() + 1, Right(), Bottom() );

	// Indicate value
	glBegin( GL_QUADS );
	{
		glColor4f( 0, 0.5, 1, 0.2 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _anchor - _min ) ) / Range(), Bottom() + 1 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _anchor - _min ) ) / Range(), Y() );
		glColor4f( 0, 0.1, 0.2, 0.2 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _value - _min ) ) / Range(), Y() );
		glVertex2f( X() + ( ( W() - 1 ) * ( _value - _min ) ) / Range(), Bottom() + 1 );
	} glEnd();
	glBegin( GL_LINES );
	{
		glColor4f( 0, 0.1, 0.2, 0.6 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _value - _min ) ) / Range() + 0.5, Y() + 1 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _value - _min ) ) / Range() + 0.5, Bottom() + 1 );
	} glEnd();
	glBegin( GL_LINES );
	{
		glColor4f( 0, 0.1, 0.2, 0.6 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _anchor - _min ) ) / Range() + 0.5, Y() + 1 );
		glVertex2f( X() + ( ( W() - 1 ) * ( _anchor - _min ) ) / Range() + 0.5, Bottom() + 1 );
	} glEnd();
	
	// Draw number selection
	if( Focused() )
	{
		float pos = X() + W() / 2.0f;
		pos += ( Label().length() * GLOut::glout.CurrentFont().W() ) / 2.0f;
		pos += ( 2 + _sel_start ) * GLOut::glout.CurrentFont().W();
		float end = pos;
		if( _sel_all )
			end += ToString( Value() ).length() * GLOut::glout.CurrentFont().W();
		glColor4f( 1, 1, 1, 0.5 );
		glRecti( pos - 2, Y() + 1, end, Bottom() - 1 );
	}
	// Draw label
	GLOut::glout << GLFormat::Align::Center()
	             << GLFormat::Location( X() + W() / 2, Y() + H() / 2 )
	             << Label() << GLFormat::Align::Left() << ": " << _value;
	
	// Draw arrows
	const unsigned char tri_left[] = { 0x08, 0x18, 0x38, 0x78, 0xf8, 0x78,
	                                   0x38, 0x18, 0x08 };
	const unsigned char tri_right[] = { 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xf0,
	                                    0xe0, 0xc0, 0x80 };

	// A lot of really useless code to show whether the new value is greater
	// than, equal to, or less than the old value.
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
										
	// Left arrow
	if( LocationOf( Value() ) - 7 >= X() )
	{
		if( !Active() )
			glColor4f( 1, 1, 1, 0.64 );
		else
		{
			if( _orig >= Value() )
				glColor4f( 1, 0.5, 0, 1 );
			else
				glColor4f( 1, 1, 1, 1 );
		}
		glRasterPos2f( LocationOf( Value() ) - 7, Y() + H() / 2 + 4 );
		glBitmap( 5, 9, 0, 0, 0, 0, tri_left );
	}
	// Right arrow
	if( LocationOf( Value() ) + 7 <= Right() )
	{
		if( !Active() )
			glColor4f( 1, 1, 1, 0.64 );
		else
		{
			if( _orig <= Value() )
				glColor4f( 1, 0.5, 0, 1 );
			else
				glColor4f( 1, 1, 1, 1 );
		}
		glRasterPos2f( LocationOf( Value() ) + 3, Y() + H() / 2 + 4 );
		glBitmap( 5, 9, 0, 0, 0, 0, tri_right );
	}
	
	// If disabled
	if( !Enabled() )
	{
		glColor4f( 0.5, 0.5, 0.5, 0.7 );
		glRecti( X(), Y(), Right() + 1, Bottom() + 1 );
	}
}

void Slider::_mouse_down()
{
	_orig = Value();
	MakeFocused();
	_sel_start = 0;
	_sel_all = true;
}

void Slider::_mouse_up()
{
	if( Value() != _orig )
		SetSignal();
	if( !Steady() )
		Value( Anchor() );
	_orig = Value();
}

void Slider::_mouse_move()
{
	if( Event::Mouse.LeftButtonDown )
	{
		if( Event::Mouse.Location.Y() > Y() + 100 ||
			Event::Mouse.Location.Y() < Y() - 100 + H() )
		{
			Value( _orig );
			ChangedSignal( _value );
		}
		else
		{
			float tmp = Value();
			float otmp = _orig;
			Value( ( ( Event::Mouse.Location.X() - Event::Mouse.DownLocation.X() +
					   LocationOf( _orig ) - X() ) * Range() ) / W() + _min );
			
			if( Event::Keyboard.ShiftDown )
				Value( ( int )( Value() / 10 + 0.5f ) * 10 );
			
			_orig = otmp;
			if( tmp != Value() )
				ChangedSignal( _value );
		}
	}
}
