/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "TextEntry.h"
#include "Text.h"
#include <GL/gl.h>

#include <iostream>
#include <sstream>

using namespace SharpConstruct::Interface;

TextEntry::TextEntry()
: _sel_start( -1 ), _sel_end( -1 )
{
	SetH( 20 );
	
	MouseMoveSignal.connect( sigc::mem_fun( *this, &TextEntry::_mouse_move ) );
	MouseDownSignal.connect( sigc::mem_fun( *this, &TextEntry::_mouse_down ) );
	MouseUpSignal.connect( sigc::mem_fun( *this, &TextEntry::_mouse_up ) );
	
	SetCursor( IBEAM );
}

TextEntry::~TextEntry()
{
	MakeUnfocused();
}

void TextEntry::KeyDown()
{
	if( Focused() )
	{		
		if( Event::Keyboard.ShiftDown && Event::Keyboard.LeftArrow )
			_sel_end--;
		else if( Event::Keyboard.ShiftDown && Event::Keyboard.RightArrow )
			_sel_end++;
		else if( Event::Keyboard.ShiftDown && Event::Keyboard.Home )
			_sel_end = 0;
		else if( Event::Keyboard.ShiftDown && Event::Keyboard.End )
			_sel_end = Label().length();
		else if( Event::Keyboard.LeftArrow )
		{
			if( _sel_start != _sel_end )
				_sel_start = std::min( _sel_start, _sel_end );
			else
				_sel_start--;
			_sel_end = _sel_start;
		}
		else if( Event::Keyboard.RightArrow )
		{
			if( _sel_start != _sel_end )
				_sel_start = std::max( _sel_start, _sel_end );
			else
				_sel_start++;
			_sel_end = _sel_start;
		}
		else if( Event::Keyboard.Enter )
		{
			ReturnSignal()();
		}
		else if( Event::Keyboard.Delete )
		{
			int st = std::min( _sel_start, _sel_end );
			int len = std::abs( _sel_end - _sel_start );
			
			if( st < Label().length() )
			{
				if( len == 0 )
					len = 1;
			
				if( st >= 0 )
				{
					std::string nv = Label();
					nv.erase( st, len );
					
					SetLabel( nv );
					
					_sel_start = _sel_end = st;
				}
			}
		}
		else if( Event::Keyboard.Backspace )
		{
			int st = std::min( _sel_start, _sel_end );
			int len = std::abs( _sel_end - _sel_start );
			
			if( len == 0 )
			{
				st--;
				len = 1;
			}
			
			if( st >= 0 )
			{
				std::string nv = Label();
				nv.erase( st, len );
				
				SetLabel( nv );
				
				_sel_start = _sel_end = st;
			}
		}
		else if( Event::Keyboard.Home )
		{
			_sel_start = _sel_end = 0;
		}
		else if( Event::Keyboard.End )
		{
			_sel_start = _sel_end = Label().length();
		}
		else if( Event::Keyboard.Key >= 0 &&
		         Event::Keyboard.Key != Event::Keyboard.VOID_KEY )
		{
			std::string nv = Label();
			nv.insert( _sel_start, &Event::Keyboard.Key );
			SetLabel( nv );
			_sel_start = _sel_end = _sel_start + 1;
		}
	}
	
	if( _sel_start < 0 )
		_sel_start = 0;
	else if( _sel_start > Label().length() )
		_sel_start = Label().length();
	
	if( _sel_end < 0 )
		_sel_end = 0;
	else if( _sel_end > Label().length() )
		_sel_end = Label().length();	
}

sigc::signal< void >& TextEntry::ReturnSignal()
{
	return _return_signal;
}

void TextEntry::_draw()
{
	// Draw border
	if( Contains() && Enabled() )
		glColor4f( 0.3, 0.3, 0.3, 1 );
	else
		glColor4f( 0.5, 0.5, 0.5, 1 );
	glRecti( X(), Y(), Right() + 1, Bottom() + 1 );
	
	 // Fill inside
	if( !Contains() || !Enabled() )
		glColor4f( 0.7, 0.7, 0.7, 1 );
	else
		glColor4f( 0.8, 0.8, 0.8, 1 );
	glRecti( X() + 1, Y() + 1, Right(), Bottom() );
	
	// Draw cursor position
	if( Focused() && _sel_start >= 0 )
	{
		glColor3f( 0.9f, 0.9f, 0.9f );
		glRecti( X() + GLOut::glout.CurrentFont().W() * _sel_start, Y() + 1,
		         X() + GLOut::glout.CurrentFont().W() * _sel_end, Bottom() - 1 );
		glColor3f( 1, 0, 0 );
		glRecti( X() + GLOut::glout.CurrentFont().W() * _sel_start, Y() + 1,
		         X() + GLOut::glout.CurrentFont().W() * _sel_start + 1, Bottom() - 1 );
	}
	
	// Draw label
	GLOut::glout << GLFormat::Align::Left()
	             << GLFormat::Location( X() + 2, Y() + H() / 2 )
	             << Label();
}

void TextEntry::_mouse_down()
{
	MakeFocused();
	_sel_start = _sel_end = ( Event::Mouse.Location.X() - X() +
	                          GLOut::glout.CurrentFont().W() / 2 ) /
	                        GLOut::glout.CurrentFont().W();
	if( _sel_start > Label().length() )
		_sel_start = _sel_end = Label().length();
}
void TextEntry::_mouse_up()
{
}
void TextEntry::_mouse_move()
{
}
