/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "MessageDialog.h"
#include "TopWindow.h"
#include "View.h"
#include "OglExt.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <cmath>
#include <iostream>

using SharpConstruct::Interface::ViewGL;

using SharpConstruct::Optimized::OPT_SIZE;
using SharpConstruct::Optimized::Point3D;
using SharpConstruct::Optimized::Normal3D;

ViewGL::ViewGL( int x, int y, int w, int h )
: Widget( x, y, w, h, "" ), _rendered( false ), _store_view( true ),
  _display_copied_view( true ), _last_mouse_location( 0, 0 ),
  _recalc_projected( true ), _only_recalc_damaged( false ),
  _first_run( true ), _initialized( false ), _texture_buffer( 0 ),
  _vertex_buffer( 0 ), _index_buffer( 0 ),
  _update_vertex_buffer( true ), _update_index_buffer( true ), _arc_ball( 0.8 ),
  _background_color( 0.3, 0.3, 0.3 ), _smoothed( true ), _double_sided( true )
{
	SetCursor( CROSS );
	MouseDownSignal.connect( sigc::mem_fun( *this, &ViewGL::_mouse_down ) );
	MouseUpSignal.connect( sigc::mem_fun( *this, &ViewGL::_mouse_up ) );
	MouseMoveSignal.connect( sigc::mem_fun( *this, &ViewGL::_mouse_move ) );
	ScrollDownSignal.connect( sigc::mem_fun( *this, &ViewGL::_scroll_down ) );
	ScrollUpSignal.connect( sigc::mem_fun( *this, &ViewGL::_scroll_up ) );
}

ViewGL::~ViewGL()
{
	if( _vertex_buffer )
		glDeleteBuffersARB( 1, &_vertex_buffer );
	if( _index_buffer )
		glDeleteBuffersARB( 1, &_index_buffer );
}

void ViewGL::ResetView()
{
	Mesh& mesh = _history->GetCurrentMesh();

	mesh._calculate_bounding_sphere();
	_location = mesh._center;
	_cam_distance = mesh._diameter;
	_arc_ball.Reset();
	_update_vertex_buffer = true;
	_recalc_projected = true;
	_rendered = false;
	_update_index_buffer = true;
}

void ViewGL::Refresh()
{
	_initialized = false;
	_update_vertex_buffer = true;
	_recalc_projected = true;
	_rendered = false;
	_update_index_buffer = true;
}

void ViewGL::SetBackgroundColor( Color input )
{
	_background_color = input;
	Refresh();
}

const SharpConstruct::Color& ViewGL::BackgroundColor() const
{
	return _background_color;
}

/*void ViewGL::UpdateTexture( Magick::Image *tex )
{
	_texture = tex;
	
	glEnable( GL_TEXTURE_2D );
	
	if( _texture_buffer )
		glDeleteTextures( 1, &_texture_buffer );

	glGenTextures( 1, &_texture_buffer );
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	glBindTexture( GL_TEXTURE_2D, _texture_buffer );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	glBindTexture( GL_TEXTURE_2D, _texture_buffer );
	
	const Magick::PixelPacket *cache =
		_texture->getConstPixels( 0, 0, _texture->columns(), _texture->rows() );

	glTexImage2D( GL_TEXTURE_2D, 0, 4, _texture->columns(),
				  _texture->rows(), 0, GL_BGRA, GL_UNSIGNED_SHORT,
				  cache );
}*/

void ViewGL::DeactivateTexture()
{
	
}

void ViewGL::ToggleSmoothed()
{
	if( _smoothed )
		_smoothed = false;
	else
		_smoothed = true;
	Refresh();
}

void ViewGL::ToggleDoubleSided()
{
	if( _double_sided )
		_double_sided = false;
	else
		_double_sided = true;
	Refresh();
}

void ViewGL::_draw_interface()
{
	_initialize();
	glClearColor( _background_color.Red, _background_color.Green,
	              _background_color.Blue, 1 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	_orientate_view();
	_draw_mesh();
}

void ViewGL::_draw()
{
	Mesh& mesh = _history->GetCurrentMesh();
	
	_make_view_context();
	
	if( !_rendered || mesh._damaged_polygon_indices.size() > 0 )
	{		
		glClear( GL_ACCUM_BUFFER_BIT );
		glDrawBuffer( GL_AUX1 );

		_draw_interface();
		_find_projected_points();
		_find_view_normals();

		_rendered = true;
	}
	glDisable( GL_SCISSOR_TEST );
	
	glDrawBuffer( GL_BACK );
	glReadBuffer( GL_AUX1 );

	_revert_view_context();
	glRasterPos2f( X(), Y() + H() );
	glCopyPixels( 0, 0, W(), H(), GL_COLOR );
	/* Debug code, draws bounding box around changed polygons
	if( _damaged_area.size() >= 4 )
	{		
		glColor3f( 1, 0, 0 );
		glBegin( GL_LINE_STRIP );
		{
			glVertex2i( _damaged_area[ 0 ] + 4, Display().H() - 24 - _damaged_area[ 1 ] );
			glVertex2i( _damaged_area[ 2 ] + 4, Display().H() - 24 - _damaged_area[ 1 ] );
			glVertex2i( _damaged_area[ 2 ] + 4, Display().H() - 24 - _damaged_area[ 3 ] );
			glVertex2i( _damaged_area[ 0 ] + 4, Display().H() - 24 - _damaged_area[ 3 ] );
			glVertex2i( _damaged_area[ 0 ] + 4, Display().H() - 24 - _damaged_area[ 1 ] );
		}
		glEnd();
	}*/
	mesh._damaged_polygon_indices.clear();
	
	/*Mesh& mesh = _history->GetCurrentMesh();
	if( mesh._damaged_polygon_indices.size() > 0 )
	{
		_only_recalc_damaged = true;
		_recalc_projected = true;
		_make_view_context();
		_orientate_view();
		_find_projected_points();
		_revert_view_context();
		
		
		glDrawBuffer( GL_AUX1 );
		
				glScissor( _damaged_area[ 0 ],
		           _damaged_area[ 1 ],
		           _damaged_area[ 2 ] - _damaged_area[ 0 ],
		           _damaged_area[ 3 ] - _damaged_area[ 1 ] );
		glEnable( GL_SCISSOR_TEST );
		//glClear( GL_DEPTH_BUFFER_BIT );
		
		// Clear depth buffer in damaged area that we're about to redraw...
		glViewport( 0, 0, Display().W(), Display().H() );
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		gluOrtho2D( Display().X(), Display().Right(),
					Display().Y(), Display().Bottom() );
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		glEnable( GL_DEPTH_TEST );
		glDepthFunc( GL_ALWAYS );
		glBegin( GL_TRIANGLES );
		glColor3f( 0, 0, 0 );
		for( std::list< unsigned >::iterator i = mesh._damaged_polygon_indices.begin();
		     i != mesh._damaged_polygon_indices.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glVertex3f( _projected_points[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							_projected_points[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							-1 );
			}
		}
		glEnd();

		_make_view_context();
		_orientate_view();
		glDepthFunc( GL_LESS );			
		
		// Find the full (view-based) damage...
		std::list< unsigned > damaged_polygon_indices;
		for( unsigned i = 0; i < _projected_points.size(); i++ )
		{
			Point3D& p = _projected_points[ i ];
			if( p.X() >= _damaged_area[ 0 ] &&
			    p.Y() >= _damaged_area[ 1 ] &&
			    p.X() <= _damaged_area[ 2 ] &&
			    p.Y() <= _damaged_area[ 3 ] )
			{
				for( unsigned j = 0; j < mesh._vertex_polygon_indices[ i ].size(); j++ )
					damaged_polygon_indices.push_back( mesh._vertex_polygon_indices[ i ][ j ] );
			}
		}
		damaged_polygon_indices.unique();
		
		Mesh& mesh = _history->GetCurrentMesh();

		
		glBegin( GL_TRIANGLES );
		for( std::list< unsigned >::iterator i = damaged_polygon_indices.begin();
		     i != damaged_polygon_indices.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glNormal3f( mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			
				glColor3f( mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Red,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Green,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Blue );
			
				glVertex3f( mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			}
		}
		glEnd();
		//glDisable( GL_SCISSOR_TEST );
		glBegin( GL_TRIANGLES );
		for( std::list< unsigned >::iterator i = mesh._damaged_polygon_indices.begin();
		     i != mesh._damaged_polygon_indices.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glNormal3f( mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			
				glColor3f( mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Red,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Green,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Blue );
			
				glVertex3f( mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			}
		}
		glEnd();
		
		mesh._damaged_polygon_indices.clear();
		_find_projected_points();
	}*/

	glDrawBuffer( GL_BACK );
	glReadBuffer( GL_BACK );
	glDisable( GL_SCISSOR_TEST );

	_revert_view_context();
}

void ViewGL::_make_view_context()
{
	glEnable( GL_COLOR_MATERIAL );
	glEnable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LESS );
	
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glViewport( 0, 0, W(), H() );
	float ratio = W();
	if( H() )
		ratio /= H();
	gluPerspective( 45, ratio, 0.1, 1000 );
}
void ViewGL::_revert_view_context()
{
	glViewport( 0, 0, TopWindow::Display().W(), TopWindow::Display().H() );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( TopWindow::Display().X(), TopWindow::Display().Right(),
	            TopWindow::Display().Bottom(), TopWindow::Display().Y() );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( -0.5, -0.5, 0 );
	
	glDisable( GL_CULL_FACE );
	glShadeModel( GL_SMOOTH );
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_COLOR_MATERIAL );
}
void ViewGL::_orientate_view()
{
	double trackball_matrix[ 16 ];
	unsigned col, row, index;

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0, 0, ( _cam_distance - _location.Z() ) * 1.5, -_location.X(),
	           -_location.Y(), _location.Z(), 0, 1, 0 );
	const ArcBall::Transform &Tmp = _arc_ball.Transformation();

	for( col = 0, index = 0; col < 4; col++ )
		for( row = 0; row < 4; row++ )
			trackball_matrix[ index++ ] = Tmp[ row ][ col ];
  
    glMultMatrixd( trackball_matrix );
}

void ViewGL::_initialize()
{
	Mesh& mesh = _history->GetCurrentMesh();
	
	if( _first_run )
	{
		if( !glexExtensionsSupported( "GL_ARB_vertex_buffer_object" ) )
		{
			const std::string msg = "Vertex Buffer Object support not found.\n"
			                        "Not all cards support this feature, but\n"
			                        "fixing the problem could be as simple as\n"
			                        "installing new drivers.";
			// There should be a message box to announce this; find a way to do
			// that gracefully later...
			std::cerr << msg << std::endl;
			exit( 1 ); // Should be safe to do this; no model can yet be loaded
		}

		GLfloat light_ambient[] = { 0.15, 0.15, 0.15, 1 };
		GLfloat light_diffuse[] = { 0.6, 0.6, 0.6, 1 };	
		GLfloat light_position_one[] = { 0, 0, -150 + _location.Z(), 0 };
		glLightfv( GL_LIGHT1, GL_AMBIENT, light_ambient );
		glLightfv( GL_LIGHT1, GL_DIFFUSE, light_diffuse );
		glLightfv( GL_LIGHT1, GL_POSITION, light_position_one );
		glEnable( GL_LIGHT1 );

		//glEnableClientState( GL_TEXTURE_COORD_ARRAY );
		glEnableClientState( GL_VERTEX_ARRAY );
		glEnableClientState( GL_NORMAL_ARRAY );
		glEnableClientState( GL_COLOR_ARRAY );
		
		_first_run = false;
	}
	
	// Useful for testing mesh functions, but should not be enabled for release:
	//glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	
	if( _smoothed )
		glShadeModel( GL_SMOOTH );
	else
		glShadeModel( GL_FLAT );	
	
	glGetDoublev( GL_PROJECTION_MATRIX, _projection_matrix );
	glGetIntegerv( GL_VIEWPORT, _viewport );

	if( _double_sided )
	{
		glDisable( GL_CULL_FACE );
		glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
	}
	else
	{
		glEnable( GL_CULL_FACE );
		glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE );
	}

	if( _vertex_buffer == 0 )
		glGenBuffersARB( 1, &_vertex_buffer );
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, _vertex_buffer );
	if( _update_vertex_buffer )
	{
		_update_vertex_buffer = false;
		/*if( mesh._modified_vertex_indices.size() == 0 )
		{*/
		glBufferDataARB( GL_ARRAY_BUFFER_ARB,
						 mesh._vertex_locations.size() * OPT_SIZE * 2 +
						 mesh._vertex_colors.size() * sizeof( Color ),
						 0, GL_STREAM_DRAW_ARB );
		// Load vertex locations...
		glBufferSubDataARB( GL_ARRAY_BUFFER_ARB, 0,
							mesh._vertex_locations.size() * OPT_SIZE,
							mesh._vertex_locations.RawData() );
		glVertexPointer( 3, GL_FLOAT, OPT_SIZE, 0 );
		// Load vertex normals...
		glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
							mesh._vertex_locations.size() * OPT_SIZE,
							mesh._vertex_normals.size() * OPT_SIZE,
							mesh._vertex_normals.RawData() );
		glNormalPointer( GL_FLOAT, OPT_SIZE,
						 ( ( char* )NULL +
						   ( mesh._vertex_locations.size() * OPT_SIZE ) ) );
		// Load vertex colors...
		glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
							mesh._vertex_locations.size() * OPT_SIZE * 2,
							mesh._vertex_colors.size() * sizeof( Color ),
							&mesh._vertex_colors[ 0 ] );
		glColorPointer( 3, GL_FLOAT, 0,
						 ( ( char* )NULL +
						   ( mesh._vertex_locations.size() * OPT_SIZE * 2 ) ) );
		/*}
		else
		{
			std::cout << mesh._modified_vertex_indices.size() << std::endl;
			
			float* p = (float*)glMapBufferARB( GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY );
			for( int i = 0; i < mesh._modified_vertex_indices.size(); i++ )
			{
				const int ndx = mesh._modified_vertex_indices[ i ];
				p[ ndx * 4 ] = mesh._vertex_locations[ ndx ].X();
				p[ ndx * 4 + 1 ] = mesh._vertex_locations[ ndx ].Y();
				p[ ndx * 4 + 2 ] = mesh._vertex_locations[ ndx ].Z();
			}
			const int s = mesh._vertex_locations.size() * 4;
			for( int i = 0; i < mesh._modified_vertex_indices.size(); i++ )
			{
				const int ndx = mesh._modified_vertex_indices[ i ];
				p[ s + ndx * 4 ] = mesh._vertex_normals[ ndx ].X();
				p[ s + ndx * 4 + 1 ] = mesh._vertex_normals[ ndx ].Y();
				p[ s + ndx * 4 + 2 ] = mesh._vertex_normals[ ndx ].Z();
			}
			glUnmapBufferARB( GL_ARRAY_BUFFER_ARB );
		}*/
					   
	}
	if( _index_buffer == 0 )
		glGenBuffersARB( 1, &_index_buffer );
	if( _update_index_buffer )
	{
		glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, _index_buffer );
		glBufferDataARB( GL_ELEMENT_ARRAY_BUFFER_ARB, mesh._indices.size() *
						 sizeof( int ), &mesh._indices[ 0 ],
						 GL_STATIC_DRAW_ARB );
		_update_index_buffer = false;
	}

	_initialized = true;
}

void ViewGL::_simple_initialize()
{
}

void ViewGL::_draw_mesh()
{
	using namespace SharpConstruct::Optimized;
	
	Mesh& mesh = _history->GetCurrentMesh();

	/*if( _update_vertex_buffer )
	{
		glBindBufferARB( GL_ARRAY_BUFFER_ARB, _vertex_buffer );
		// Update vertex locations
		glBufferSubDataARB( GL_ARRAY_BUFFER_ARB, mesh._lowest_index *
		                    OPT_SIZE, ( mesh._highest_index -
                            mesh._lowest_index + 1 ) * OPT_SIZE,
		                    mesh._vertex_locations[ mesh._lowest_index ].RawData() );
		// Update vertex normals
		glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
		                    ( mesh._vertex_locations.size() +
		                      mesh._lowest_index ) * OPT_SIZE,
		                    ( mesh._highest_index -
                              mesh._lowest_index + 1 ) * OPT_SIZE,
		                    mesh._vertex_normals[ mesh._lowest_index ].RawData() );
		// Update vertex colors
		glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
		                    mesh._vertex_locations.size() * OPT_SIZE * 2 +
		                    mesh._lowest_index * sizeof( Color ),
		                    ( mesh._highest_index -
                              mesh._lowest_index + 1 ) * sizeof( Color ),
		                    &mesh._vertex_colors[ mesh._lowest_index ] );

		_update_vertex_buffer = false;
	}*/
	
	//glBindTexture( GL_TEXTURE_2D, _texture_buffer );
	
	/*std::cout << "Total: " << mesh._vertex_locations.size()
	          << ", Damaged: " << mesh._modified_vertex_indices.size()
	          << ", Damaged Range: " << mesh._highest_index - mesh._lowest_index << std::endl;*/
	
	glDrawRangeElements( GL_TRIANGLES, 0, mesh._indices.size() - 1,
	                     mesh._indices.size(), GL_UNSIGNED_INT, 0 );
	/*glBegin( GL_TRIANGLES );
	for( unsigned i = 0; i < mesh._indices.size(); i++ )
	{
		glNormal3f( mesh._vertex_normals[ mesh._indices[ i ] ].X(),
		            mesh._vertex_normals[ mesh._indices[ i ] ].Y(),
		            mesh._vertex_normals[ mesh._indices[ i ] ].Z() );
		glColor3f( mesh._vertex_colors[ mesh._indices[ i ] ].Red,
		           mesh._vertex_colors[ mesh._indices[ i ] ].Green,
		           mesh._vertex_colors[ mesh._indices[ i ] ].Blue );
		glVertex3f( mesh._vertex_locations[ mesh._indices[ i ] ].X(),
		            mesh._vertex_locations[ mesh._indices[ i ] ].Y(),
		            mesh._vertex_locations[ mesh._indices[ i ] ].Z() );
	}
	glEnd();*/
}

void ViewGL::_find_view_normals()
{
	GLdouble model_matrix[ 16 ];
	double projected_x, projected_y, projected_z;
	double zero_loc[ 3 ];
	
	glGetDoublev( GL_MODELVIEW_MATRIX, model_matrix );
	
	gluUnProject( 0, 0, 0, model_matrix, _projection_matrix, _viewport,
		        &zero_loc[ 0 ], &zero_loc[ 1 ], &zero_loc[ 2 ] );

	// Update Mesh::_view_right_normal
	gluUnProject( 1, 0, 0, model_matrix, _projection_matrix, _viewport,
	            &projected_x, &projected_y, &projected_z );
	Mesh::_view_right_normal = Normal3D( projected_x, projected_y, projected_z );
	Mesh::_view_right_normal.X() -= zero_loc[ 0 ];
	Mesh::_view_right_normal.Y() -= zero_loc[ 1 ];
	Mesh::_view_right_normal.Z() -= zero_loc[ 2 ];
	Mesh::_view_right_normal.Normalize();
	
	// Update Mesh::_view_up_normal
	gluUnProject( 0, -1, 0, model_matrix, _projection_matrix, _viewport,
	            &projected_x, &projected_y, &projected_z );
	Mesh::_view_up_normal = Normal3D( projected_x, projected_y, projected_z );
	Mesh::_view_up_normal.X() -= zero_loc[ 0 ];
	Mesh::_view_up_normal.Y() -= zero_loc[ 1 ];
	Mesh::_view_up_normal.Z() -= zero_loc[ 2 ];
	Mesh::_view_up_normal.Normalize();
}

void ViewGL::_find_projected_points()
{
	GLdouble model_matrix[ 16 ];
	double projected_x, projected_y, projected_z;
	Mesh& mesh = _history->GetCurrentMesh();

	glGetDoublev( GL_MODELVIEW_MATRIX, model_matrix );
	
	float depth;
	
	glReadPixels( Event::Mouse.Location.X() - X(),
	              H() - ( Event::Mouse.Location.Y() - Y() ), 1, 1,
	              GL_DEPTH_COMPONENT, GL_FLOAT, &depth );
	
	gluUnProject( Event::Mouse.Location.X() - X(),
	              H() - ( Event::Mouse.Location.Y() - Y() ), depth,
	              model_matrix, _projection_matrix, _viewport, &projected_x,
	              &projected_y, &projected_z );

	mesh._current_selection = Point3D( projected_x, projected_y, projected_z );
	
	//if( _recalc_projected ) // Maybe someday...
	/*{
		std::cout << "recalcing proj..." << std::endl;
		if( _projected_points.size() != mesh._vertex_locations.size() || !_only_recalc_damaged )
		{
			std::cout << "recalcing all..." << std::endl;
			_projected_points.resize( mesh._vertex_locations.size() );
			for( int i = 0; i < mesh._vertex_locations.size(); i++ )
			{
				gluProject( mesh._vertex_locations[ i ].X(),
							mesh._vertex_locations[ i ].Y(),
							mesh._vertex_locations[ i ].Z(),
							model_matrix, _projection_matrix, _viewport,
							&projected_x, &projected_y, &projected_z );
				_projected_points[ i ] = Point3D( projected_x, projected_y, projected_z );
			}
		}
		if( _only_recalc_damaged )
		{
			std::cout << "recalcing damaged..." << std::endl;
			for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
			{
				gluProject( mesh._vertex_locations[ mesh._modified_vertex_indices[ i ] ].X(),
							mesh._vertex_locations[ mesh._modified_vertex_indices[ i ] ].Y(),
							mesh._vertex_locations[ mesh._modified_vertex_indices[ i ] ].Z(),
							model_matrix, _projection_matrix, _viewport,
							&projected_x, &projected_y, &projected_z );
				_projected_points[ mesh._modified_vertex_indices[ i ] ] = Point3D( projected_x, projected_y, projected_z );
			}
			
			_only_recalc_damaged = false;
		}
		else
			std::cout << "NO RECALC!!!" << std::endl;
		_recalc_projected = false;
	}*/
}

void ViewGL::_mouse_down()
{
	if( Event::Mouse.LeftButtonDown /*&& !Fl::event_shift()*/ )
	{
		_make_view_context();
		_orientate_view();
		_draw_mesh();
		_find_projected_points();
		_revert_view_context();
		_history->AddMesh(); // Bottleneck!
		if( MainBrush->GetModelingMode() == Brush::GRAB )
		{
			//_simple_initialize();
			//_initialize();
			//_find_view_normals();
			//_find_projected_points();
			_history->GetCurrentMesh().StartGrab();
		}
		ChangedMeshSignal();
		
	}
	else if( Event::Mouse.MiddleButtonDown )
	{
		ResetView();
		_rendered = false;
	}
	else if( Event::Mouse.RightButtonDown )
	{
		double xnorm = _last_mouse_location.X();
		double ynorm = _last_mouse_location.Y();
		ArcBall::NormalizeCoordinates( xnorm, ynorm, W(), H() );
		_arc_ball.BeginDrag( xnorm, ynorm );
	}
}

void ViewGL::_mouse_up()
{
	_history->GetCurrentMesh().EndGrab();
	_store_view = true;
	_display_copied_view = true;
	_rendered = false;
	_recalc_projected = true;
}

void ViewGL::_mouse_move()
{
	Mesh& mesh = _history->GetCurrentMesh();
	
	//_simple_initialize();
	//_initialize();
	//_find_view_normals();
	//_find_projected_points();
	if( Event::Keyboard.AltDown )
	{
		//_location.X() += ( Fl::event_x() - _x_last ) / 100.0;
		//_location.Y() += ( H() - Fl::event_y() - 1 - _y_last ) / 100.0;
		
		_location.X() += ( Event::Mouse.Location.X() - _last_mouse_location.X() ) / 100.0f;
		_location.Y() += ( Event::Mouse.Location.Y() - _last_mouse_location.Y() ) / -100.0f;
		_rendered = false;
	}
	else if( Event::Mouse.LeftButtonDown )
	{
		_update_vertex_buffer = true;
		
		bool changed = false;
		if( Event::Keyboard.ShiftDown )
		{
			MainBrush->SetStrength( -MainBrush->GetStrength() );
			changed = true;
		}
		//mesh._modified_vertex_indices.clear();
		if( MainBrush->GetModelingMode() == Brush::DISPLACE )
		{
			_history->GetCurrentMesh().DisplaceAreas();
			
			/*for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
			{
				_damaged_polygon_indices.insert( _damaged_polygon_indices.begin(),
					mesh._vertex_polygon_indices[
					mesh._modified_vertex_indices[ i ] ].begin(),
					mesh._vertex_polygon_indices[
					mesh._modified_vertex_indices[ i ] ].end() );
			}
			_damaged_polygon_indices.unique();*/
			/*if( mesh._modified_vertex_indices.size() > 0 )
			{//x1,y1,x2,y2
				_damaged_area.resize( 4 );
				_damaged_area[ 0 ] = W();
				_damaged_area[ 1 ] = H();
				_damaged_area[ 2 ] = 0;
				_damaged_area[ 3 ] = 0;
				for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); ++i )
				{
					Point3D& p = _projected_points[ mesh._modified_vertex_indices[ i ] ];
					if( p.X() < _damaged_area[ 0 ] )
						_damaged_area[ 0 ] = p.X();
					if( p.Y() < _damaged_area[ 1 ] )
						_damaged_area[ 1 ] = p.Y();
					if( p.X() > _damaged_area[ 2 ] )
						_damaged_area[ 2 ] = p.X();
					if( p.Y() > _damaged_area[ 3 ] )
						_damaged_area[ 3 ] = p.Y();
				}
				
			}*/
			//_rendered = false;
			
			//if( MainBrush->GetColorEdit() )
			/*if( _texture != NULL )
				UpdateTexture( _texture );*/
		}
		else if( MainBrush->GetModelingMode() == Brush::GRAB )
		{
			_history->GetCurrentMesh().ContinueGrab( ( Event::Mouse.Location.X() -
														_last_mouse_location.X() ) / 1000.0,
													 ( Event::Mouse.Location.Y() -
														_last_mouse_location.Y() ) / 1000.0 );
			_rendered = false;
		}
		else if( MainBrush->GetModelingMode() == Brush::SMOOTH )
			_history->GetCurrentMesh().SmoothAreas();
		
		if( changed )
			MainBrush->SetStrength( -MainBrush->GetStrength() );
	}
	else if( Event::Mouse.RightButtonDown )
	{
		double xnorm = _last_mouse_location.X();
		double ynorm = _last_mouse_location.Y();
		ArcBall::NormalizeCoordinates( xnorm, ynorm, W(), H() );
		_arc_ball.Drag( xnorm, ynorm );
		_rendered = false;
		_store_view = false;
		_display_copied_view = false;
	}
	_last_mouse_location = Event::Mouse.Location;
}

void ViewGL::_scroll_down()
{
	_location.Z() -= _cam_distance / 10;
	_rendered = false;
}

void ViewGL::_scroll_up()
{
	_location.Z() += _cam_distance / 10;
	_rendered = false;
}
