/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "TopWindow.h"
#include "XDisplayGL.h"
#include <GL/gl.h>
#include <X11/keysym.h>
#include <ctime>

#include <iostream>

using namespace SharpConstruct::Interface;

XLib::Display* XDisplayGL::_xdisplay;

static float down_time;

XDisplayGL::XDisplayGL( TopWindow& ctrl, int w, int h, std::string lbl )
: _controller( ctrl ), _label( lbl ), _cursor( Widget::DEFAULT ),
  _check_double_click( true )
{
	_initialize( w, h );
}

XDisplayGL::~XDisplayGL()
{
	if( _glx_context )
	{
		if( !glXMakeCurrent( _xdisplay, None, NULL ) )
			fprintf( stderr, "Error releasing drawing context\n" );
		glXDestroyContext( _xdisplay, _glx_context );
	}
}

void XDisplayGL::Draw()
{
	//_make_current();
}

void XDisplayGL::Postwork()
{	
	glXSwapBuffers( _xdisplay, _window );
}

int XDisplayGL::W() const
{
	XWindowAttributes tmp;
	XGetWindowAttributes( _xdisplay, _window, &tmp );
	return tmp.width;
}
int XDisplayGL::H() const
{
	XWindowAttributes tmp;
	XGetWindowAttributes( _xdisplay, _window, &tmp );
	return tmp.height;
}
void XDisplayGL::SetSize( const int w, const int h )
{
	XResizeWindow( _xdisplay, _window, w, h );
}
std::string XDisplayGL::Label() const
{
	return _label;
}
void XDisplayGL::SetLabel( const std::string& lbl )
{
	_label = lbl;
}

void XDisplayGL::SetCursor( Widget::Cursors c )
{
	if( _cursor != c )
	{
		_cursor = c;
		XDefineCursor( _xdisplay, _window, XCreateFontCursor( _xdisplay, c ) );
		glXWaitX(); // Makes sure X updates the cursor
	}
}

void XDisplayGL::_initialize( int w, int h )
{
	XVisualInfo *vi;
	Colormap cmap;
	Atom wmDelete;

	_xdisplay = XOpenDisplay( NULL );
	
	_screen = DefaultScreen( _xdisplay );

	static int attrListDbl[] =
	{
		GLX_RGBA, GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, 4,
		GLX_GREEN_SIZE, 4,
		GLX_BLUE_SIZE, 4,
		GLX_DEPTH_SIZE, 16,
		None
	};

	vi = glXChooseVisual( _xdisplay, _screen, attrListDbl );

	_glx_context = glXCreateContext( _xdisplay, vi, 0, GL_TRUE );

	cmap = XCreateColormap( _xdisplay, RootWindow( _xdisplay, vi->screen ),
	                        vi->visual, AllocNone );
	
	_attributes.colormap = cmap;

	_attributes.event_mask =
		ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask |
		ButtonReleaseMask | StructureNotifyMask | PointerMotionMask;

	_window = XCreateWindow( _xdisplay, RootWindow( _xdisplay, vi->screen ),
	                         0, 0, w, h,
	                         0, vi->depth, InputOutput, vi->visual,
	                         CWColormap | CWEventMask,
	                         &_attributes );
	wmDelete = XInternAtom( _xdisplay, "WM_DELETE_WINDOW", True );
	XSetWMProtocols( _xdisplay, _window, &wmDelete, 1 );
	XSetStandardProperties( _xdisplay, _window, Label().c_str(),
	                        Label().c_str(), None, NULL, 0, NULL );
	XMapRaised( _xdisplay, _window );

	_make_current();
}

void XDisplayGL::_make_current()
{
	glXMakeCurrent( _xdisplay, _window, _glx_context );
}

void XDisplayGL::Run()
{

}

void XDisplayGL::DoEvents()
{	
	bool DoMouseMove = false;
	XEvent event;
	
	// Causes grab/drag bug...
	/*if( Event::Mouse.AnyButtonDown() )
	{
		const float c = (float)clock() / CLOCKS_PER_SEC;if( c - down_time > MouseData::MOUSE_DOWN_REPEAT_DELAY )
		{
			_controller.MouseStillDown();
			down_time = c;
		}
	}*/
	
	while( XEventsQueued( _xdisplay, QueuedAfterReading ) > 0 )
	{
		XNextEvent( _xdisplay, &event );
		
		ResponderWidgets::Responders.Clear();
		ResponderDialogs::Responders.Clear();
		
		switch( event.type )
		{
		case Expose:
			if( event.xexpose.count != 0 )
				break;
			//std::cout << "expose" << std::endl;
			break;
		case ConfigureNotify:
			/*if( ( event.xconfigure.width != W() ) ||
				( event.xconfigure.height != H() ) )
			{
				SetW( event.xconfigure.width );
				SetH( event.xconfigure.height );*/
				_controller.Resize();
			//}
			//std::cout << "configurenotify" << std::endl;
			break;
		case ClientMessage:
			if( *XGetAtomName( _xdisplay, event.xclient.message_type ) ==
				*"WM_PROTOCOLS" )
				Close();
			//std::cout << "clientmessage" << std::endl;
			break;
			case ButtonPress:
			{
				bool dc = (float)clock() / CLOCKS_PER_SEC < down_time + 0.3;
				down_time = (float)clock() / CLOCKS_PER_SEC;
	
				if( event.xbutton.button == 4 )
					_controller.ScrollUp();
				else if( event.xbutton.button == 5 )
					_controller.ScrollDown();
				else
				{
					if( event.xbutton.button == 1 )
						Event::Mouse.LeftButtonDown = true;
					else if( event.xbutton.button == 2 )
						Event::Mouse.MiddleButtonDown = true;
					else if( event.xbutton.button == 3 )
						Event::Mouse.RightButtonDown = true;
					else if( event.xbutton.button == 6 )
					{
						Event::Keyboard.ControlDown = true;
						Event::Keyboard.Key = 'z';
						_controller.KeyDown();
						Event::Mouse.S1ButtonDown = true;
					}
					else if( event.xbutton.button == 7 )
					{
						Event::Keyboard.ControlDown = true;
						Event::Keyboard.ShiftDown = true;
						Event::Keyboard.Key = 'z';
						_controller.KeyDown();
						Event::Mouse.S2ButtonDown = true;
					}
					Event::Mouse.DownLocation = Event::Mouse.Location;
					
					_controller.MouseDown();
					if( dc && _check_double_click )
					{
						_controller.DoubleClick();
						_check_double_click = false;
					}
					else
					{
						_check_double_click = true;
					}
				}
				//std::cout << "press" << std::endl;
				break;
			}
		case ButtonRelease:
			if( event.xbutton.button == 1 )
				Event::Mouse.LeftButtonDown = false;
			else if( event.xbutton.button == 2 )
				Event::Mouse.MiddleButtonDown = false;
			else if( event.xbutton.button == 3 )
				Event::Mouse.RightButtonDown = false;
			else if( event.xbutton.button == 6 )
			{
				Event::Keyboard.ControlDown = false;
				Event::Mouse.S1ButtonDown = false;
			}
			else if( event.xbutton.button == 7 )
			{
				Event::Keyboard.ControlDown = false;
				Event::Keyboard.ShiftDown = false;
				Event::Mouse.S2ButtonDown = false;
			}
			_controller.MouseUp();
			//std::cout << "release" << std::endl;
			break;
		case MotionNotify:
			Event::Mouse.Location.SetX( event.xmotion.x );
			Event::Mouse.Location.SetY( event.xmotion.y );
			DoMouseMove = true;
			//std::cout << "motion" << std::endl;
			break;
		case KeyPress:
			KeySym key_sym;
			XLookupString( &event.xkey, 0, 0, &key_sym, 0 );
			Event::Keyboard.Key = XK_VoidSymbol;

			/* I'd be grateful to anyone who could tell me an easy way to work
			   around this X11 foolishness of having to specify two versions of
			   the same key, the main key plus the "KP" version.
			*/
		
			if( key_sym >= XK_KP_0 && key_sym <= XK_KP_9 )
				key_sym -= 0xff80;
		
			switch( key_sym )
			{
			case XK_Shift_L:
			case XK_Shift_R:
				Event::Keyboard.ShiftDown = true;
				break;
			case XK_Control_L:
			case XK_Control_R:
				Event::Keyboard.ControlDown = true;
				break;
			case XK_Alt_L:
			case XK_Alt_R:
				Event::Keyboard.AltDown = true;
				break;
			case XK_Left:
				Event::Keyboard.LeftArrow = true;
				_controller.KeyDown();
				break;
			case XK_Right:
				Event::Keyboard.RightArrow = true;
				_controller.KeyDown();
				break;
			case XK_Return:
			case XK_KP_Enter:
				Event::Keyboard.Enter = true;
				_controller.KeyDown();
				break;
			case XK_Delete:
			case XK_KP_Delete:
				Event::Keyboard.Delete = true;
				_controller.KeyDown();
				break;
			case XK_BackSpace:
				Event::Keyboard.Backspace = true;
				_controller.KeyDown();
				break;
			case XK_Home:
				Event::Keyboard.Home = true;
				_controller.KeyDown();
				break;
			case XK_End:
				Event::Keyboard.End = true;
				_controller.KeyDown();
				break;
			case XK_Tab:
				Event::Keyboard.Tab = true;
				_controller.KeyDown();
				break;
			default:
				Event::Keyboard.Key = key_sym;
				_controller.KeyDown();
				break;
			}
			break;
		case KeyRelease:
			{
				KeySym key_sym;
				XLookupString( &event.xkey, 0, 0, &key_sym, 0 );
				Event::Keyboard.Key = XK_VoidSymbol;
			
				switch( key_sym )
				{
				case XK_Shift_L:
				case XK_Shift_R:
					Event::Keyboard.ShiftDown = false;
					break;
				case XK_Control_L:
				case XK_Control_R:
					Event::Keyboard.ControlDown = false;
					break;
				case XK_Alt_L:
				case XK_Alt_R:
					Event::Keyboard.AltDown = false;
					break;
				case XK_Left:
					Event::Keyboard.LeftArrow = false;
					break;
				case XK_Right:
					Event::Keyboard.RightArrow = false;
					break;
				case XK_Return:
				case XK_KP_Enter:
					Event::Keyboard.Enter = false;
					break;
				case XK_Delete:
				case XK_KP_Delete:
					Event::Keyboard.Delete = false;
					break;
				case XK_BackSpace:
					Event::Keyboard.Backspace = false;
					break;
				case XK_Home:
					Event::Keyboard.Home = false;
					break;
				case XK_End:
					Event::Keyboard.End = false;
					break;
				case XK_Tab:
					Event::Keyboard.Tab = false;
					break;
				default:
					break;
				}
				break;
			}
		}
	}
	
	/* Big thanks to the FLTK source code for showing me this trick;
	   by waiting to process the move events the interface is hugely
	   more responsive, especially with a very complex mesh loaded.*/
	if( DoMouseMove )
		_controller.MouseMove();
}

void XDisplayGL::Close()
{
	if( _controller.Close() )
	{
		if( _glx_context )
		{
			if( !glXMakeCurrent( _xdisplay, None, NULL ) ) {
				fprintf( stderr, "Error releasing drawing context\n" );
			}
			glXDestroyContext( _xdisplay, _glx_context );
		}
		exit( 0 );
	}
}
