use ExtUtils::MakeMaker;

my @exe_files = ('bin/intclock');
my $pm_name   = "Intclock";
my $configfile = "lib/Intclock.pl";
my @configLines = ();

my $datadir   = "/usr/share/intclock";
push @configLines, "\$config_datadir = \"$datadir\";"; 

my @graphicModules = qw(Gtk2 Gtk);
my %supportedGraphicModules = ();
my $preferred;

for (my $i = 0 ; $i < scalar(@ARGV) ; $i++){
  if ($ARGV[$i] =~ /^GUI=(.+)$/){
    my $gui = $1;
    if (not grep /^$gui$/, @graphicModules) {
      die "Unsupported GUI option: $gui\n";
    }
    $preferred = $gui;
    splice(@ARGV, $i, 1);
    $i--;
  }
}

#------------------------------------------------------------------------
# preliminary checks
#------------------------------------------------------------------------
# Mandatory perl modules
foreach my $module qw(Data::Dumper Time::HiRes Locale::Maketext) {
  print "Checking whether $module perl module is available... ";
  eval "use $module";
  if ($@) {
    print "no\n";
    die "Perl module $module is required for this program!\n";
  }
  else {
    print "yes\n";
  }
}

my $default_gui;
# Choice between Gtk and Gtk2
foreach my $module (@graphicModules) {
  print "Checking whether $module perl module is available... ";
  eval "use $module";
  if ($@) {
    print "no\n";
  }
  else {
    push @supportedGraphicModules, $module;
    print "yes\n";
  }
}

if (not @supportedGraphicModules) {
  die "At least one of the following perl modules is required: @graphicModules!\n";
}

if (defined $preferred) {
  if (grep /^$preferred$/, @supportedGraphicModules) {
    $default_gui = $preferred;
  }
  else {
    print "WARNING: Requested graphic module '$preferred' not available\n";
    $default_gui = shift @supportedGraphicModules;
    print "Using '$default_gui' instead\n";
  }
}
else {
  $default_gui = shift @supportedGraphicModules;
  if ($default_gui ne "Gtk2") {
    print "**************************************************************\n";
    print " NOTE: The Gtk2 Perl module was not found.\n\n";
    print " This is not a problem, because the program can also work with\n";
    print " the Gtk module.  However, the preferred graphic interface\n";
    print " is via Gtk2 (the newer features in intclock 2.x are not ported\n";
    print " to the Gtk interface), so consider installing the perl-Gtk2\n";
    print " Perl module anyway: see http://gtk2-perl.sourceforge.net.\n\n";
    print " If you do, you will have to execute 'perl Makefile.PL' again\n";
    print " afterwards.\n";
    print "**************************************************************\n";
  }
}

push @configLines, "\@config_guis = qw(@supportedGraphicModules);";
push @configLines, "\$config_default_gui = \"$default_gui\";";

# Availability of Encode::_utf8_on
print "Checking whether the Encode::_utf8_on perl function is available... ";
my $encode_utf8_supported;
eval {
  use Encode;
  my $string = "test";
  Encode::_utf8_on($string);
  return 0;
};
if ($@) {
  $encode_utf8_supported = 0;
  print "no\n";
}
else {
  $encode_utf8_supported = 1;
  print "yes\n";
}
push @configLines, "\$config_utf8_on = $encode_utf8_supported;";

# UTF8-correctness of Data::Dumper
if ($encode_utf8_supported) {
  print "Checking whether Data::Dumper module works correctly with utf8... ";
  eval {
    use Data::Dumper;
    use Encode;
    my $var = chr(0x0100);
    my $d = new Data::Dumper([$var]);
    my $output = $d->Dump();
    die "UTF8 flag not honored"
	if not (Encode::is_utf8($output) or $output =~ /\\x/);
  };
  if ($@) {
    push @configLines, "\$config_dd_fix_utf8 = 1;";
    print "no\n";
  }
  else {
    push @configLines, "\$config_dd_fix_utf8 = 0;";
    print "yes\n";
  }
  
  print "Checking whether POSIX::strftime works correctly with utf8... ";
  eval {
    use locale;
    use POSIX qw(locale_h strftime);
    setlocale(LC_ALL, "el_GR.UTF-8");
    my $str = strftime("%a", 0, 0, 12, 3, 3, 105, 0, 92, 1);
    die "UTF8 flag not honored"
	if not (Encode::is_utf8($str));
  };
  if ($@) {
    push @configLines, "\$config_strftime_fix_utf8 = 1;";
    print "no\n";
  }
  else {
    push @configLines, "\$config_strftime_fix_utf8 = 0;";
    print "yes\n";
  }
}

# Write everything to config file
print "Writing configuration file $configfile\n";
open CONFIG, ">$configfile" or die "Can't write to $configfile: $!\n";
print CONFIG "# Configuration created by Makefile.PL\n";
print CONFIG "BEGIN {\n";
foreach my $line (@configLines) {
  print CONFIG "  $line\n";
}
print CONFIG "}\n";
print CONFIG "1;\n";
close CONFIG;

#------------------------------------------------------------------------
# writing Makefile
#------------------------------------------------------------------------

sub MY::postamble {
  my $section = q{
DATADIR = $datadir

install:: extra_install

extra_install::
	mkdir -p $(DATADIR)
	PERL5LIB=blib/lib perl data/mk_desktop_entry data/intclock.desktop /usr/share/applications/intclock.desktop
	cp data/intclock.png $(DATADIR)
	cp data/territories.dat $(DATADIR)

all:: syntax_check

syntax_check::
	$(NOECHO) for file in $(TO_INST_PM) $(EXE_FILES); do \
          if [ $$file -nt .$@ ]; then \
            PERL5LIB=blib/lib perl -c $$file; \
	  fi \
	done
	$(NOECHO) $(TOUCH) .$@
  };
  $section =~ s/\$datadir/$datadir/gs;
  return $section;
}

WriteMakefile(NAME          => $pm_name,
	      DISTNAME      => 'intclock',
	      VERSION_FROM  => 'bin/intclock',
	      EXE_FILES     => \@exe_files,
	      NO_META       => "true",
	      );
